@@ -356,9 +356,10 @@ bool roca_check
 	return false;
 }
 
-void rsa_check
+int rsa_check
 	(mpz_srcptr rsa_n, mpz_srcptr rsa_e)
 {
+	int ret = 0;
 	std::cout << "Public-key algorithm: " << std::endl << "\tRSA" << std::endl;
 	std::cout << "Security level of public key: " << std::endl << "\t";
 	std::cout << "|n| = " << mpz_sizeinbase(rsa_n, 2L) << " bit, ";
@@ -369,7 +370,10 @@ void rsa_check
 	std::cout << std::endl << "\t";
 	std::cout << "n ";
 	if (mpz_probab_prime_p(rsa_n, TMCG_MR_ITERATIONS))
+	{
 		std::cout << "IS PROBABLE PRIME" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
 	else
 		std::cout << "is not probable prime" << std::endl << "\t";
 	for (size_t i = 1; i < 4; i++)
@@ -382,7 +386,10 @@ void rsa_check
 	for (size_t i = 0; i < PRIMES_SIZE; i++)
 	{
 		if (mpz_divisible_ui_p(rsa_n, primes[i]))
+		{
 			std::cout << primes[i] << " * ";
+			ret = -3; // return with non-null status code
+		}
 	}
 	std::cout << "..." << std::endl << "\t";
 	std::cout << "Legendre-Jacobi symbol (i/n): ";
@@ -402,22 +409,126 @@ void rsa_check
 	}
 	std::cout << "#(+1) = " << pos << " #(-1) = " << neg << std::endl << "\t";
 	if (roca_check(rsa_n))
+	{
 		std::cout << "n is SUSPICIOUS for the ROCA vulnerability" <<
 			std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "e is ";
 	if (!mpz_probab_prime_p(rsa_e, TMCG_MR_ITERATIONS))
+	{
 		std::cout << "NOT ";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "probable prime" << std::endl << "\t";
 	std::cout << "e is ";
 	if (mpz_cmp_ui(rsa_e, 41) < 0)
+	{
 		std::cout << "VERY SMALL" << std::endl;
+		ret = -3; // return with non-null status code
+	}
 	else
 		std::cout << "okay" << std::endl;
+	return ret;
 }
 
-void dsa_check
+int rsa_check
+	(mpz_srcptr rsa_p, mpz_srcptr rsa_q, mpz_srcptr rsa_d, mpz_srcptr rsa_n, mpz_srcptr rsa_e)
+{
+	int ret = 0;
+	size_t nbits = mpz_sizeinbase(rsa_n, 2L);
+	mpz_t pm1, qm1, tmp, foo, bar;
+	std::cout << "\t";
+	if (!mpz_probab_prime_p(rsa_p, TMCG_MR_ITERATIONS))
+	{
+		std::cout << "p IS NOT PROBABLE PRIME" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	if (!mpz_probab_prime_p(rsa_q, TMCG_MR_ITERATIONS))
+	{
+		std::cout << "q IS NOT PROBABLE PRIME" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_init_set(pm1, rsa_p), mpz_init_set(qm1, rsa_q); 
+	mpz_sub_ui(pm1, pm1, 1L), mpz_sub_ui(qm1, qm1, 1L);
+	mpz_init(tmp);
+	mpz_gcd(tmp, pm1, rsa_e);
+	if (mpz_cmp_ui(tmp, 1L))
+	{
+		std::cout << "p and e ARE NOT RELATIVELY PRIME" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_gcd(tmp, qm1, rsa_e);
+	if (mpz_cmp_ui(tmp, 1L))
+	{
+		std::cout << "q and e ARE NOT RELATIVELY PRIME" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_init(foo), mpz_init(bar);
+	mpz_mul(foo, rsa_p, rsa_q);
+	if (mpz_cmp(foo, rsa_n) != 0)
+	{
+		std::cout << "n IS NOT p * q" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_sub(foo, rsa_q, rsa_p);
+	mpz_abs(foo, foo);
+	mpz_set_ui(bar, 1L);
+	mpz_mul_2exp(bar, bar, (nbits / 2) - 100);
+	if (mpz_cmp(foo, bar) <= 0)
+	{
+		std::cout << "|p - q| IS NOT > 2^(nlen/2 - 100)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_set_ui(bar, 1L);
+	mpz_mul_2exp(bar, bar, (nbits / 2) - 1);
+	if (mpz_cmp(rsa_p, bar) <= 0)
+	{
+		std::cout << "p IS NOT > 2^(nlen/2 - 1)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	if (mpz_cmp(rsa_q, bar) <= 0)
+	{
+		std::cout << "q IS NOT > 2^(nlen/2 - 1)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_set_ui(bar, 1L);
+	mpz_mul_2exp(bar, bar, (nbits / 2));
+	if (mpz_cmp(rsa_d, bar) <= 0)
+	{
+		std::cout << "d IS NOT > 2^(nlen/2)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_lcm(tmp, pm1, qm1);
+	if (mpz_cmp(rsa_d, tmp) >= 0)
+	{
+		std::cout << "d IS NOT < lcm(p-1,q-1)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	if (!mpz_invert(foo, rsa_e, tmp))
+	{
+		std::cout << "e IS NOT INVERTIBLE mod lcm(p-1,q-1)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	if (mpz_cmp(foo, rsa_d))
+	{
+		std::cout << "d IS NOT e^(-1) mod lcm(p-1,q-1)" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
+	mpz_clear(foo), mpz_clear(bar);
+	mpz_clear(tmp), mpz_clear(pm1), mpz_clear(qm1);
+	if (ret == 0)
+		std::cout << "RSA private key -- no conspicuousness found";
+	else
+		std::cout << "RSA private key -- CONSPICUOUSNESS found";
+	std::cout << std::endl;	
+	return ret;
+}
+
+int dsa_check
 	(mpz_srcptr dsa_p, mpz_srcptr dsa_q, mpz_srcptr dsa_g, mpz_srcptr dsa_y)
 {
+	int ret = 0;
 	std::cout << "Public-key algorithm: " << std::endl << "\tDSA" << std::endl;
 	std::cout << "Security level of DSA domain parameter set: " <<
 		std::endl << "\t";
@@ -427,7 +538,10 @@ void dsa_check
 	std::cout << std::endl << "\t";
 	std::cout << "p is ";
 	if (!mpz_probab_prime_p(dsa_p, TMCG_MR_ITERATIONS))
+	{
 		std::cout << "NOT ";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "probable prime" << std::endl << "\t";
 	mpz_t pm1;
 	mpz_init_set(pm1, dsa_p);
@@ -449,7 +563,10 @@ void dsa_check
 	std::cout << std::endl << "\t";
 	std::cout << "q is ";
 	if (!mpz_probab_prime_p(dsa_q, TMCG_MR_ITERATIONS))
+	{
 		std::cout << "NOT ";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "probable prime" << std::endl << "\t";
 	mpz_set(pm1, dsa_q);
 	mpz_sub_ui(pm1, pm1, 1L);
@@ -467,7 +584,10 @@ void dsa_check
 	std::cout << "g is ";
 	mpz_powm(pm1, dsa_g, dsa_q, dsa_p);
 	if (mpz_cmp_ui(pm1, 1L))
+	{
 		std::cout << "NOT ";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "generator of G_q" << std::endl << "\t";
 	mpz_t tmp, foo, bar;
 	mpz_init(tmp), mpz_init_set_ui(foo, 2L), mpz_init(bar);
@@ -485,7 +605,10 @@ void dsa_check
 	std::cout << "y is ";
 	mpz_powm(pm1, dsa_y, dsa_q, dsa_p);
 	if (mpz_cmp_ui(pm1, 1L))
+	{
 		std::cout << "NOT ";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "element of G_q" << std::endl << "\t";
 	bool trivial = false;
 	for (size_t i = 0; i < TRIVIAL_SIZE; i++)
@@ -507,17 +630,22 @@ void dsa_check
 	if (!trivial)
 		std::cout << "y is not trivial" << std::endl << "\t";
 	else
+	{
 		std::cout << "y is TRIVIAL, i.e., y = g^c mod p (for some |c| < " <<
 			TRIVIAL_SIZE << ")" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "Legendre-Jacobi symbol (y/p) is " <<
 		mpz_jacobi(dsa_y, dsa_p) << std::endl;
 	mpz_clear(pm1);
 	mpz_clear(tmp);
+	return ret;
 }
 
-void elg_check
+int elg_check
 	(mpz_srcptr elg_p, mpz_srcptr elg_g, mpz_srcptr elg_y, mpz_srcptr dsa_q)
 {
+	int ret = 0;
 	mpz_t pm1, tmp, bar;
 	std::cout << "Public-key algorithm: " << std::endl <<
 		"\tElGamal" << std::endl;
@@ -528,7 +656,10 @@ void elg_check
 		std::endl << "\t";
 	std::cout << "p is ";
 	if (!mpz_probab_prime_p(elg_p, TMCG_MR_ITERATIONS))
+	{
 		std::cout << "NOT ";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "probable prime" << std::endl << "\t";
 	mpz_init_set(pm1, elg_p);
 	mpz_sub_ui(pm1, pm1, 1L);
@@ -568,8 +699,11 @@ void elg_check
 		mpz_ui_pow_ui(bar, it->first, it->second);
 		mpz_powm(tmp, elg_g, bar, elg_p);
 		if (!mpz_cmp_ui(tmp, 1L))
+		{
 			std::cout << "is VERY SMALL (" << it->first << "^" <<
 				it->second << " elements) ";
+			ret = -3; // return with non-null status code
+		}
 		else
 			std::cout << "is okay ";
 	}
@@ -605,25 +739,36 @@ void elg_check
 	if (!trivial)
 		std::cout << "y is not trivial" << std::endl << "\t";
 	else
+	{
 		std::cout << "y is TRIVIAL, i.e., y = g^c mod p (for some |c| < " <<
 			TRIVIAL_SIZE << ")" << std::endl << "\t";
+		ret = -3; // return with non-null status code
+	}
 	std::cout << "Legendre-Jacobi symbol (y/p) is " <<
 		mpz_jacobi(elg_y, elg_p) << std::endl;
 	mpz_clear(pm1);
+	return ret;
 }
 
-void sig_check_dsa
+int sig_check_dsa
 	(mpz_srcptr dsa_p, mpz_srcptr dsa_q, mpz_srcptr dsa_g, mpz_srcptr dsa_y,
 	 mpz_srcptr dsa_r)
 {
+	int ret = 0;
 	if (!mpz_cmp_ui(dsa_r, 1L))
+	{
 		std::cout << "r is WEAK (i.e. k = 0)" << std::endl << "\t";
+		ret = -4; // return with non-null status code
+	}
 	mpz_t pm1, tmp;
 	bool suspicious = false;
 	mpz_init(pm1), mpz_init(tmp);
 	mpz_mod(pm1, dsa_y, dsa_q);
 	if (!mpz_cmp(dsa_r, pm1))
+	{
 		std::cout << "r is WEAK (i.e. k = x)" << std::endl << "\t";
+		ret = -4; // return with non-null status code
+	}
 	for (size_t i = 0; i < TRIVIAL_SIZE; i++)
 	{
 		mpz_powm_ui(tmp, dsa_g, i, dsa_p);
@@ -641,8 +786,12 @@ void sig_check_dsa
 		}
 	}
 	if (suspicious)
+	{
 		std::cout << "r is SUSPICIOUS (small k used)" << std::endl << "\t";
+		ret = -4; // return with non-null status code
+	}
 	mpz_clear(pm1), mpz_clear(tmp);
+	return ret;
 }
 
 int main
@@ -652,10 +801,11 @@ int main
 	static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 	static const char *version = PACKAGE_VERSION " (" PACKAGE_NAME ")";
 
-	std::string	ifilename, kfilename, rfilename;
+	std::string	kfilename, filename, ofilename;
 	int			opt_verbose = 0;
-	bool		opt_binary = false, opt_reduce = false, opt_private = false;
+	bool		opt_b = false, opt_r = false, opt_p = false, opt_y = false;
 	char		*opt_k = NULL;
+	char		*opt_o = NULL;
 
 	// parse command line arguments
 	for (size_t i = 0; i < (size_t)(argc - 1); i++)
@@ -667,15 +817,25 @@ int main
 			size_t idx = ++i;
 			if ((idx < (size_t)(argc - 1)) && (opt_k == NULL))
 			{
-				rfilename = argv[i+1];
-				opt_k = (char*)rfilename.c_str();
+				kfilename = argv[i+1];
+				opt_k = (char*)kfilename.c_str();
+			}
+			continue;
+		}
+		else if ((arg.find("-o") == 0))
+		{
+			size_t idx = ++i;
+			if ((idx < (size_t)(argc - 1)) && (opt_o == NULL))
+			{
+				ofilename = argv[i+1];
+				opt_o = (char*)ofilename.c_str();
 			}
 			continue;
 		}
 		else if ((arg.find("--") == 0) || (arg.find("-b") == 0) || 
 		    (arg.find("-r") == 0) || (arg.find("-v") == 0) ||
 		    (arg.find("-h") == 0) || (arg.find("-V") == 0) ||
-			(arg.find("-p") == 0))
+			(arg.find("-p") == 0) || (arg.find("-y") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -688,6 +848,8 @@ int main
 				std::cout << "  -h, --help     print this help" << std::endl;
 				std::cout << "  -k FILENAME    use keyring FILENAME" <<
 					" containing external revocation keys" << std::endl;
+				std::cout << "  -o FILENAME    export (reduced) public key" << 
+					" to FILENAME" << std::endl;
 				std::cout << "  -p, --private  read from private key block" <<
 					std::endl;
 				std::cout << "  -r, --reduce   check only valid subkeys" <<
@@ -696,14 +858,18 @@ int main
 					std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" <<
 					std::endl;
+				std::cout << "  -y, --yaot     list keys from keyring" <<
+					" FILENAME with user ID containing KEYFILE" << std::endl;
 				return 0; // not continue
 			}
 			if ((arg.find("-b") == 0) || (arg.find("--binary") == 0))
-				opt_binary = true;
+				opt_b = true;
 			if ((arg.find("-p") == 0) || (arg.find("--private") == 0))
-				opt_private = true;
+				opt_p = true;
 			if ((arg.find("-r") == 0) || (arg.find("--reduce") == 0))
-				opt_reduce = true;
+				opt_r = true;
+			if ((arg.find("-y") == 0) || (arg.find("--yaot") == 0))
+				opt_y = true;
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
 			{
 				std::cout << "dkg-keycheck v" << version << std::endl;
@@ -718,19 +884,51 @@ int main
 			std::cerr << "ERROR: unknown option \"" << arg << "\"" << std::endl;
 			return -1;
 		}
-		kfilename = arg;
+		filename = arg;
 	}
 
+#ifdef DKGPG_TESTSUITE_Y
+	if (tmcg_mpz_wrandom_ui() % 2)
+	{
+		filename = "TestY-pub.asc";
+	}
+	else
+	{
+		filename = "TestY-sec.asc";
+		opt_p = true;
+	}
+	opt_verbose = 2;
+#endif
+
 	// check command line arguments
-	if (kfilename.length() == 0)
+	if (!opt_y && filename.length() == 0)
 	{
 		std::cerr << "ERROR: argument KEYFILE is missing; usage: " <<
 			usage << std::endl;
 		return -1;
 	}
+	if (opt_y && (opt_k == NULL))
+	{
+		std::cerr << "ERROR: no keyring specified (option -k missing)" <<
+			std::endl;
+		return -1;
+	} 
+
+	// lock memory
+	bool force_secmem = false;
+	if (opt_p)
+	{
+		if (!lock_memory())
+		{
+			std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK" <<
+				" required for full memory protection" << std::endl;
+			// at least try to use libgcrypt's secure memory
+			force_secmem = true;
+		}
+	}
 
 	// initialize LibTMCG
-	if (!init_libTMCG())
+	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
 		return -1;
@@ -740,36 +938,39 @@ int main
 			version_libTMCG() << std::endl;
 
 	// choose input format of key file
-	tmcg_openpgp_armor_t kformat = TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK; 
-	if (opt_private)
-		kformat = TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK;
+	tmcg_openpgp_armor_t format = TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK; 
+	if (opt_p)
+		format = TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK;
 
 	// read the key file
 	std::string armored_pubkey;
-	if (opt_binary)
+	if (!opt_y)
 	{
-		if (!read_binary_key_file(kfilename, kformat, armored_pubkey))
-			return -1;
-	}
-	else
-	{
-		if (!read_key_file(kfilename, armored_pubkey))
-			return -1;
-	}
-
-	// read the keyring
-	tmcg_openpgp_armor_t rformat = TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK;
-	std::string armored_pubring;
-	if (opt_k)
-	{
-		if (opt_binary)
+		if (opt_b)
 		{
-			if (!read_binary_key_file(rfilename, rformat, armored_pubring))
+			if (!read_binary_key_file(filename, format, armored_pubkey))
 				return -1;
 		}
 		else
 		{
-			if (!read_key_file(rfilename, armored_pubring))
+			if (!read_key_file(filename, armored_pubkey))
+				return -1;
+		}
+	}
+
+	// read the keyring
+	tmcg_openpgp_armor_t kformat = TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK;
+	std::string armored_pubring;
+	if (opt_k != NULL)
+	{
+		if (opt_b)
+		{
+			if (!read_binary_key_file(kfilename, kformat, armored_pubring))
+				return -1;
+		}
+		else
+		{
+			if (!read_key_file(kfilename, armored_pubring))
 				return -1;
 		}
 	}
@@ -779,7 +980,7 @@ int main
 	TMCG_OpenPGP_Pubkey *primary = NULL;
 	TMCG_OpenPGP_Keyring *ring = NULL;
 	bool parse_ok;
-	if (opt_k)
+	if (opt_k != NULL)
 	{
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
 			PublicKeyringParse(armored_pubring, opt_verbose, ring);
@@ -791,20 +992,32 @@ int main
 	}
 	else
 		ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
-	if (opt_private)
+	if (opt_y)
 	{
-		std::string passphrase = ""; // first try an empty passphrase
+		int ret = 0;
+		if (ring->List(filename) == 0)
+			ret = -1;
+		delete ring;
+		return ret;
+	}
+	else if (opt_p)
+	{
+		tmcg_openpgp_secure_string_t passphrase = ""; // try an empty passphrase
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
 			PrivateKeyBlockParse(armored_pubkey, opt_verbose, passphrase, prv);
 		if (!parse_ok)
 		{
-			if (!get_passphrase("Enter passphrase to unlock private key",
+#ifdef DKGPG_TESTSUITE_Y
+			passphrase = "TestY";
+#else
+			if (!get_passphrase("Enter passphrase to unlock private key", false,
 				passphrase))
 			{
 				std::cerr << "ERROR: cannot read passphrase" << std::endl;
 				delete ring;
 				return -1;
 			}
+#endif
 			parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
 				PrivateKeyBlockParse(armored_pubkey, opt_verbose, passphrase,
 				prv);
@@ -815,17 +1028,31 @@ int main
 			PublicKeyBlockParse(armored_pubkey, opt_verbose, primary);
 	if (parse_ok)
 	{
-		if (opt_private)
+		if (opt_p)
 		{
 			primary = prv->pub; // get the public key from private key
 			prv->RelinkPublicSubkeys(); // relink the contained subkeys
 		}
 		primary->CheckSelfSignatures(ring, opt_verbose);
 		primary->CheckSubkeys(ring, opt_verbose);
-		if (opt_reduce)
+		if (opt_r)
 			primary->Reduce();
-		if (primary->weak(opt_verbose) && opt_verbose)
+		if (primary->Weak(opt_verbose) && opt_verbose)
 			std::cerr << "WARNING: weak primary key detected" << std::endl;
+		if (opt_o != NULL)
+		{
+			tmcg_openpgp_octets_t pub;
+			std::string armor;
+			primary->Export(pub);
+			CallasDonnerhackeFinneyShawThayerRFC4880::
+				ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, pub, armor);
+			if (!write_message(ofilename, armor))
+			{
+				delete primary;
+				delete ring;
+				return -1;
+			} 
+		}
 	}
 	else
 	{
@@ -835,11 +1062,12 @@ int main
 	}
 
 	// show information w.r.t. primary key
+	int ret = 0;
 	std::string kid, fpr;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		KeyidCompute(primary->pub_hashing, kid);
 	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(primary->pub_hashing, fpr);
+		FingerprintComputePretty(primary->pub_hashing, fpr);
 	std::cout << "OpenPGP V4 Key ID of primary key: " << std::endl << "\t";
 	std::cout << kid << std::endl;
 	std::cout << "OpenPGP V4 fingerprint of primary key: " << std::endl << "\t";
@@ -854,7 +1082,10 @@ int main
 		// compute validity period of the primary key after key creation time
 		time_t ekeytime = primary->creationtime + primary->expirationtime;
 		if (ekeytime < time(NULL))
+		{
 			std::cout << "[EXPIRED] ";
+			ret = -2; // return with non-null status code
+		}
 		std::cout << ctime(&ekeytime);
 	}
 	std::cout << "OpenPGP Revocation Keys: " << std::endl;
@@ -864,7 +1095,7 @@ int main
 		tmcg_openpgp_octets_t f(rk.key_fingerprint,
 			 rk.key_fingerprint+sizeof(rk.key_fingerprint));
 		CallasDonnerhackeFinneyShawThayerRFC4880::
-			FingerprintConvert(f, fpr);
+			FingerprintConvertPretty(f, fpr);
 		std::cout << "\t" << fpr << std::endl;
 	}
 	if (primary->revkeys.size() == 0)
@@ -908,14 +1139,14 @@ int main
 		{
 			std::cerr << "ERROR: cannot convert RSA key material" << std::endl;
 			mpz_clear(rsa_n), mpz_clear(rsa_e);
-			if (opt_private)
+			if (opt_p)
 				delete prv;
 			else
 				delete primary;
 			delete ring;
 			return -1;
 		}
-		rsa_check(rsa_n, rsa_e);
+		ret = rsa_check(rsa_n, rsa_e);
 		mpz_clear(rsa_n), mpz_clear(rsa_e);
 	}
 	else if (primary->pkalgo == TMCG_OPENPGP_PKALGO_DSA)
@@ -931,14 +1162,14 @@ int main
 			std::cerr << "ERROR: cannot convert DSA key material" << std::endl;
 			mpz_clear(dsa_p), mpz_clear(dsa_q), mpz_clear(dsa_g);
 			mpz_clear(dsa_y), mpz_clear(dsa_r);
-			if (opt_private)
+			if (opt_p)
 				delete prv;
 			else
 				delete primary;
 			delete ring;
 			return -1;
 		}
-		dsa_check(dsa_p, dsa_q, dsa_g, dsa_y);
+		ret = dsa_check(dsa_p, dsa_q, dsa_g, dsa_y);
 		// additional part for checking DSA signatures
 		for (size_t i = 0; i < primary->selfsigs.size(); i++)
 		{
@@ -951,7 +1182,7 @@ int main
 				}
 				else
 				{
-					sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+					ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 					sigs.push_back(primary->selfsigs[i]);
 				}
 			}
@@ -970,7 +1201,7 @@ int main
 				}
 				else
 				{
-					sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+					ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 					sigs.push_back(primary->keyrevsigs[i]);
 				}
 			}
@@ -989,7 +1220,7 @@ int main
 				}
 				else
 				{
-					sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+					ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 					sigs.push_back(primary->certrevsigs[i]);
 				}
 			}
@@ -1011,7 +1242,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sig);
 					}
 				}
@@ -1031,7 +1262,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sig);
 					}
 				}
@@ -1056,7 +1287,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sig);
 					}
 				}
@@ -1076,7 +1307,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sig);
 					}
 				}
@@ -1088,32 +1319,98 @@ int main
 		mpz_clear(dsa_p), mpz_clear(dsa_q), mpz_clear(dsa_g);
 		mpz_clear(dsa_y), mpz_clear(dsa_r);
 	}
+	else if (primary->pkalgo == TMCG_OPENPGP_PKALGO_ECDSA)
+	{
+		unsigned int curvebits = 0;
+		const char *curvename = gcry_pk_get_curve(primary->key, 0, &curvebits);
+		if (curvename == NULL)
+			curvename = "unknown";
+		std::cout << "Public-key algorithm: " << std::endl <<
+			"\tECDSA with curve \"" << curvename << "\" with " <<
+			curvebits << " bits" << std::endl;
+	}
+	else if (primary->pkalgo == TMCG_OPENPGP_PKALGO_EDDSA)
+	{
+		unsigned int curvebits = 0;
+		const char *curvename = gcry_pk_get_curve(primary->key, 0, &curvebits);
+		if (curvename == NULL)
+			curvename = "unknown";
+		std::cout << "Public-key algorithm: " << std::endl <<
+			"\tEdDSA with curve \"" << curvename << "\" with " <<
+			curvebits << " bits" << std::endl;
+	}
 	else
 	{
 		std::cerr << "ERROR: public-key algorithm not supported" << std::endl;
-		if (opt_private)
+		if (opt_p)
 			delete prv;
 		else
 			delete primary;
 		delete ring;
 		return -1;
 	}
+	if (opt_p)
+	{
+		if (prv->Weak(opt_verbose) && opt_verbose)
+		{
+			std::cerr << "WARNING: weak private primary key detected" <<
+				std::endl;
+		}
+		switch (prv->pkalgo)
+		{
+			case TMCG_OPENPGP_PKALGO_RSA:
+			case TMCG_OPENPGP_PKALGO_RSA_ENCRYPT_ONLY:
+			case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
+				{
+					mpz_t rsa_p, rsa_q, rsa_d, rsa_n, rsa_e;
+					mpz_init(rsa_p), mpz_init(rsa_q), mpz_init(rsa_d);
+					mpz_init(rsa_n), mpz_init(rsa_e);
+					if (!tmcg_mpz_set_gcry_mpi(prv->rsa_p, rsa_p) ||
+						!tmcg_mpz_set_gcry_mpi(prv->rsa_q, rsa_q) ||
+						!tmcg_mpz_set_gcry_mpi(prv->rsa_d, rsa_d) ||
+						!tmcg_mpz_set_gcry_mpi(primary->rsa_n, rsa_n) ||
+						!tmcg_mpz_set_gcry_mpi(primary->rsa_e, rsa_e))
+					{
+						std::cerr << "ERROR: cannot convert RSA key material" <<
+							std::endl;
+						mpz_clear(rsa_p), mpz_clear(rsa_q), mpz_clear(rsa_d);
+						mpz_clear(rsa_n), mpz_clear(rsa_e);
+						delete prv;
+						delete ring;
+						return -1;
+					}
+					ret = rsa_check(rsa_p, rsa_q, rsa_d, rsa_n, rsa_e);
+					mpz_clear(rsa_p), mpz_clear(rsa_q), mpz_clear(rsa_d);
+					mpz_clear(rsa_n), mpz_clear(rsa_e);
+				}
+				break;
+			case TMCG_OPENPGP_PKALGO_DSA:
+			case TMCG_OPENPGP_PKALGO_ECDSA:
+			case TMCG_OPENPGP_PKALGO_EDDSA:
+				break;
+			default:
+				break;
+		}
+	}
 	// show information w.r.t. (valid) user IDs
 	for (size_t j = 0; j < primary->userids.size(); j++)
 	{
 		std::cout << "OpenPGP User ID: " << std::endl << "\t";
-		std::cout << primary->userids[j]->userid_sanitized << std::endl;
+		std::cout << primary->userids[j]->userid_sanitized;
+		if (!primary->userids[j]->valid)
+			std::cout << " [INVALID]";
+		std::cout << std::endl;
 	}
 	// show information w.r.t. (valid) subkeys
 	for (size_t j = 0; j < primary->subkeys.size(); j++)
 	{
 		TMCG_OpenPGP_Subkey *sub = primary->subkeys[j];
-		if (sub->weak(opt_verbose) && opt_verbose)
+		if (sub->Weak(opt_verbose) && opt_verbose)
 			std::cerr << "WARNING: weak subkey detected" << std::endl;
 		CallasDonnerhackeFinneyShawThayerRFC4880::
 			KeyidCompute(sub->sub_hashing, kid);
 		CallasDonnerhackeFinneyShawThayerRFC4880::
-			FingerprintCompute(sub->sub_hashing, fpr);
+			FingerprintComputePretty(sub->sub_hashing, fpr);
 		std::cout << "OpenPGP V4 Key ID of subkey: " << std::endl << "\t";
 		std::cout << kid << std::endl;
 		std::cout << "OpenPGP V4 fingerprint of subkey: " << std::endl << "\t";
@@ -1125,7 +1422,7 @@ int main
 			std::cout << "undefined" << std::endl;
 		else
 		{
-			// compute validity period of the primary key after key creation
+			// compute validity period of the subkey after key creation
 			time_t ekeytime = sub->creationtime + sub->expirationtime;
 			if (ekeytime < time(NULL))
 				std::cout << "[EXPIRED] ";
@@ -1138,7 +1435,7 @@ int main
 			tmcg_openpgp_octets_t f(rk.key_fingerprint,
 				 rk.key_fingerprint+sizeof(rk.key_fingerprint));
 			CallasDonnerhackeFinneyShawThayerRFC4880::
-				FingerprintConvert(f, fpr);
+				FingerprintConvertPretty(f, fpr);
 			std::cout << "\t" << fpr << std::endl;
 		}
 		if (sub->revkeys.size() == 0)
@@ -1183,14 +1480,14 @@ int main
 				std::cerr << "ERROR: cannot convert RSA key material" <<
 					std::endl;
 				mpz_clear(rsa_n), mpz_clear(rsa_e);
-				if (opt_private)
+				if (opt_p)
 					delete prv;
 				else
 					delete primary;
 				delete ring;
 				return -1;
 			}
-			rsa_check(rsa_n, rsa_e);
+			ret = rsa_check(rsa_n, rsa_e);
 			mpz_clear(rsa_n), mpz_clear(rsa_e);
 		}
 		else if (sub->pkalgo == TMCG_OPENPGP_PKALGO_ELGAMAL)
@@ -1206,7 +1503,7 @@ int main
 					std::endl;
 				mpz_clear(elg_p), mpz_clear(elg_g), mpz_clear(elg_y);
 				mpz_clear(dsa_q);
-				if (opt_private)
+				if (opt_p)
 					delete prv;
 				else
 					delete primary;
@@ -1221,7 +1518,7 @@ int main
 						std::endl;
 					mpz_clear(elg_p), mpz_clear(elg_g), mpz_clear(elg_y);
 					mpz_clear(dsa_q);
-					if (opt_private)
+					if (opt_p)
 						delete prv;
 					else
 						delete primary;
@@ -1231,7 +1528,7 @@ int main
 			}
 			else
 				mpz_set_ui(dsa_q, 0L);
-			elg_check(elg_p, elg_g, elg_y, dsa_q);
+			ret = elg_check(elg_p, elg_g, elg_y, dsa_q);
 			mpz_clear(elg_p), mpz_clear(elg_g), mpz_clear(elg_y);
 			mpz_clear(dsa_q);
 		}
@@ -1249,22 +1546,52 @@ int main
 					std::endl;
 				mpz_clear(dsa_p), mpz_clear(dsa_q), mpz_clear(dsa_g);
 				mpz_clear(dsa_y);
-				if (opt_private)
+				if (opt_p)
 					delete prv;
 				else
 					delete primary;
 				delete ring;
 				return -1;
 			}
-			dsa_check(dsa_p, dsa_q, dsa_g, dsa_y);
+			ret = dsa_check(dsa_p, dsa_q, dsa_g, dsa_y);
 			mpz_clear(dsa_p), mpz_clear(dsa_q), mpz_clear(dsa_g);
 			mpz_clear(dsa_y);
 		}
+		else if (sub->pkalgo == TMCG_OPENPGP_PKALGO_ECDH)
+		{
+			unsigned int curvebits = 0;
+			const char *curvename = gcry_pk_get_curve(sub->key, 0, &curvebits);
+			if (curvename == NULL)
+				curvename = "unknown";
+			std::cout << "Public-key algorithm: " << std::endl <<
+				"\tECDH with curve \"" << curvename << "\" with " <<
+				curvebits << " bits" << std::endl;
+		}
+		else if (sub->pkalgo == TMCG_OPENPGP_PKALGO_ECDSA)
+		{
+			unsigned int curvebits = 0;
+			const char *curvename = gcry_pk_get_curve(sub->key, 0, &curvebits);
+			if (curvename == NULL)
+				curvename = "unknown";
+			std::cout << "Public-key algorithm: " << std::endl <<
+				"\tECDSA with curve \"" << curvename << "\" with " <<
+				curvebits << " bits" << std::endl;
+		}
+		else if (sub->pkalgo == TMCG_OPENPGP_PKALGO_EDDSA)
+		{
+			unsigned int curvebits = 0;
+			const char *curvename = gcry_pk_get_curve(sub->key, 0, &curvebits);
+			if (curvename == NULL)
+				curvename = "unknown";
+			std::cout << "Public-key algorithm: " << std::endl <<
+				"\tEdDSA with curve \"" << curvename << "\" with " <<
+				curvebits << " bits" << std::endl;
+		}
 		else
 		{
 			std::cerr << "ERROR: public-key algorithm not supported" <<
 				std::endl;
-			if (opt_private)
+			if (opt_p)
 				delete prv;
 			else
 				delete primary;
@@ -1286,7 +1613,7 @@ int main
 					std::endl;
 				mpz_clear(dsa_p), mpz_clear(dsa_q), mpz_clear(dsa_g);
 				mpz_clear(dsa_y), mpz_clear(dsa_r);
-				if (opt_private)
+				if (opt_p)
 					delete prv;
 				else
 					delete primary;
@@ -1304,7 +1631,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sub->selfsigs[i]);
 					}
 				}
@@ -1323,7 +1650,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sub->bindsigs[i]);
 					}
 				}
@@ -1342,7 +1669,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sub->keyrevsigs[i]);
 					}
 				}
@@ -1361,7 +1688,7 @@ int main
 					}
 					else
 					{
-						sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
+						ret = sig_check_dsa(dsa_p, dsa_q, dsa_g, dsa_y, dsa_r);
 						sigs.push_back(sub->certrevsigs[i]);
 					}
 				}
@@ -1373,6 +1700,54 @@ int main
 			mpz_clear(dsa_y), mpz_clear(dsa_r);
 		}
 	}
+	if (opt_p)
+	{
+		prv->RelinkPrivateSubkeys(); // undo the relinking
+		for (size_t j = 0; j < prv->private_subkeys.size(); j++)
+		{
+			TMCG_OpenPGP_PrivateSubkey *ssb = prv->private_subkeys[j];
+			if (ssb->Weak(opt_verbose) && opt_verbose)
+				std::cerr << "WARNING: weak private subkey #" << j <<
+					" detected" << std::endl;
+			switch (ssb->pkalgo)
+			{
+				case TMCG_OPENPGP_PKALGO_RSA:
+				case TMCG_OPENPGP_PKALGO_RSA_ENCRYPT_ONLY:
+				case TMCG_OPENPGP_PKALGO_RSA_SIGN_ONLY:
+					{
+						mpz_t rsa_p, rsa_q, rsa_d, rsa_n, rsa_e;
+						mpz_init(rsa_p), mpz_init(rsa_q), mpz_init(rsa_d);
+						mpz_init(rsa_n), mpz_init(rsa_e);
+						if (!tmcg_mpz_set_gcry_mpi(ssb->rsa_p, rsa_p) ||
+							!tmcg_mpz_set_gcry_mpi(ssb->rsa_q, rsa_q) ||
+							!tmcg_mpz_set_gcry_mpi(ssb->rsa_d, rsa_d) ||
+							!tmcg_mpz_set_gcry_mpi(ssb->pub->rsa_n, rsa_n) ||
+							!tmcg_mpz_set_gcry_mpi(ssb->pub->rsa_e, rsa_e))
+						{
+							std::cerr << "ERROR: cannot convert RSA key" <<
+								" material" << std::endl;
+							mpz_clear(rsa_p), mpz_clear(rsa_q), mpz_clear(rsa_d);
+							mpz_clear(rsa_n), mpz_clear(rsa_e);
+							delete prv;
+							delete ring;
+							return -1;
+						}
+						ret = rsa_check(rsa_p, rsa_q, rsa_d, rsa_n, rsa_e);
+						mpz_clear(rsa_p), mpz_clear(rsa_q), mpz_clear(rsa_d);
+						mpz_clear(rsa_n), mpz_clear(rsa_e);
+					}
+					break;
+				case TMCG_OPENPGP_PKALGO_ELGAMAL:
+				case TMCG_OPENPGP_PKALGO_DSA:
+				case TMCG_OPENPGP_PKALGO_ECDSA:
+				case TMCG_OPENPGP_PKALGO_EDDSA:
+				case TMCG_OPENPGP_PKALGO_ECDH:
+					break;
+				default:
+					break;
+			}
+		}
+	}
 
 	// cross-check DSA signatures w.r.t same k
 	std::cout << std::endl;
@@ -1387,17 +1762,18 @@ int main
 			{
 				std::cout << "WEAKNESS: DSA value r is EQUAL for both" <<
 					" signatures (e.g. same k used)" << std::endl;
+				ret = -2; // return with non-null status code
 			}
 		}
 	}
 
 	// release primary key and keyring structures
-	if (opt_private)
+	if (opt_p)
 		delete prv;
 	else
 		delete primary;
 	delete ring;
 	
-	return 0;
+	return ret;
 }
 
