@@ -59,17 +59,18 @@ std::vector<std::string>	peers;
 bool						instance_forked = false;
 
 tmcg_openpgp_secure_string_t	passphrase;
-std::string						kfilename;
+std::string						kfilename, reason;
 std::string						passwords, hostname, port;
 
 int 							opt_verbose = 0;
 char							*opt_passwords = NULL;
 char							*opt_hostname = NULL;
 char							*opt_k = NULL;
+char							*opt_R = NULL;
 unsigned long int				opt_r = 0, opt_p = 55000, opt_W = 5;
 
 void run_instance
-	(size_t whoami, const time_t sigtime, const size_t reason,
+	(size_t whoami, const time_t sigtime, const size_t rcode,
 	 const size_t num_xtests)
 {
 	// read the key file
@@ -139,30 +140,28 @@ void run_instance
 		delete ring;
 		exit(-1);
 	}
+	delete ring;
 	if (!prv->pub->valid || prv->Weak(opt_verbose))
 	{
 		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	if (prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7)
 	{
 		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 	TMCG_OpenPGP_PrivateSubkey *sub = NULL;
 	if (prv->private_subkeys.size())
-		sub = prv->private_subkeys[0];
+		sub = prv->private_subkeys[0]; // FIXME: only the first subkey is used
 
 	// create an instance of tDSS by stored parameters from private key
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = NULL;
 	if (!init_tDSS(prv, opt_verbose, dss))
 	{
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -171,7 +170,6 @@ void run_instance
 	{
 		std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" << std::endl;
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
@@ -187,10 +185,9 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read" <<
-					" password for protecting channel to R_" << i << std::endl;
+				std::cerr << "ERROR: p_" << whoami << ": " << "cannot read" <<
+					" password for protecting channel to p_" << i << std::endl;
 				delete dss;
-				delete ring;
 				delete prv;
 				exit(-1);
 			}
@@ -198,11 +195,10 @@ void run_instance
 			if (((i + 1) < peers.size()) &&
 				!TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip" <<
-					" to next password for protecting channel to R_" <<
+				std::cerr << "ERROR: p_" << whoami << ": " << "cannot skip" <<
+					" to next password for protecting channel to p_" <<
 					(i + 1) << std::endl;
 				delete dss;
-				delete ring;
 				delete prv;
 				exit(-1);
 			}
@@ -210,7 +206,7 @@ void run_instance
 		else
 		{
 			// use simple key -- we assume that GNUnet provides secure channels
-			key << "dkg-revoke::R_" << (i + whoami);
+			key << "dkg-revoke::p_" << (i + whoami);
 		}
 		uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
@@ -244,68 +240,9 @@ void run_instance
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
-	for (size_t i = 0; i < num_xtests; i++)
-	{
-		mpz_t xtest;
-		mpz_init_set_ui(xtest, i);
-		std::cerr << "INFO: R_" << whoami << ": xtest = " << xtest << " <-> ";
-		rbc->Broadcast(xtest);
-		for (size_t ii = 0; ii < peers.size(); ii++)
-		{
-			if (!rbc->DeliverFrom(xtest, ii))
-				std::cerr << "<X> ";
-			else
-				std::cerr << xtest << " ";
-		}
-		std::cerr << std::endl;
-		mpz_clear(xtest);
-	}
-
+	xtest(num_xtests, whoami, peers.size(), rbc);
 	// participants must agree on a common signature creation time (OpenPGP)
-	if (opt_verbose)
-		std::cerr << "INFO: agree on a signature creation time for OpenPGP" <<
-			std::endl;
-	time_t csigtime = 0;
-	std::vector<time_t> tvs;
-	mpz_t mtv;
-	mpz_init_set_ui(mtv, sigtime);
-	rbc->Broadcast(mtv);
-	tvs.push_back(sigtime);
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		if (i != whoami)
-		{
-			if (rbc->DeliverFrom(mtv, i))
-			{
-				time_t utv;
-				utv = (time_t)mpz_get_ui(mtv);
-				tvs.push_back(utv);
-			}
-			else
-			{
-				std::cerr << "WARNING: R_" << whoami << ": no signature" <<
-					" creation time stamp received from " << i << std::endl;
-			}
-		}
-	}
-	mpz_clear(mtv);
-	std::sort(tvs.begin(), tvs.end());
-	if (tvs.size() < (peers.size() - T_RBC))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": not enough timestamps" <<
-			" received" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	// use a median value as some kind of gentle agreement
-	csigtime = tvs[tvs.size()/2];
-	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": canonicalized signature" <<
-			" creation time = " << csigtime << std::endl;
-
+	time_t csigtime = agree_time(sigtime, whoami, peers.size(), opt_verbose, rbc);
 	// select hash algorithm for OpenPGP based on |q| (size in bit)
 	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
 	if (mpz_sizeinbase(dss->q, 2L) == 256)
@@ -316,330 +253,111 @@ void run_instance
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 	else
 	{
-		std::cerr << "ERROR: R_" << whoami << ": selecting hash algorithm" <<
+		std::cerr << "ERROR: p_" << whoami << ": selecting hash algorithm" <<
 			" failed for |q| = " << mpz_sizeinbase(dss->q, 2L) << std::endl;
 		delete rbc, delete aiou, delete aiou2;
 		delete dss;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 
-	// reason for revocation
-	tmcg_openpgp_byte_t revcode = 0x00;
-	if (reason < 255)
-		revcode = reason;
-
-	// compute a hash of pub and sub, respectively
-	tmcg_openpgp_octets_t trailer_pub, pub_hashing, hash_pub, left_pub,
-		trailer_sub, sub_hashing, hash_sub, left_sub, issuer;
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		FingerprintCompute(prv->pub->pub_hashing, issuer);
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		PacketSigPrepareRevocationSignature(TMCG_OPENPGP_SIGNATURE_KEY_REVOCATION,
-			hashalgo, csigtime, revcode, "", issuer, trailer_pub);
-	if (sub != NULL)
+	// machine-readable code that denotes the reason for the revocation
+	tmcg_openpgp_revcode_t revcode = TMCG_OPENPGP_REVCODE_NO_REASON_SPECIFIED;
+	if (rcode < 255)
 	{
-		CallasDonnerhackeFinneyShawThayerRFC4880::
-			PacketSigPrepareRevocationSignature(TMCG_OPENPGP_SIGNATURE_SUBKEY_REVOCATION,
-				hashalgo, csigtime, revcode, "", issuer, trailer_sub);
+		revcode = (tmcg_openpgp_revcode_t)rcode;
 	}
+	else
+	{
+		std::cerr << "WARNING: machine-readable revocation code is" <<
+				" too large and ignored" << std::endl;
+	}
+	switch (revcode)
+	{
+		case TMCG_OPENPGP_REVCODE_NO_REASON_SPECIFIED:
+			break;
+		case TMCG_OPENPGP_REVCODE_KEY_SUPERSEDED:
+			if (opt_verbose)
+			{
+				std::cerr << "INFO: machine-readable revocation code means" <<
+					" that key is superseded" << std::endl;
+			}
+			break;
+		case TMCG_OPENPGP_REVCODE_KEY_COMPROMISED:
+			if (opt_verbose)
+			{
+				std::cerr << "INFO: machine-readablerevocation code means" <<
+					" that key material has been compromised" << std::endl;
+			}
+			break;
+		case TMCG_OPENPGP_REVCODE_KEY_RETIRED:
+			if (opt_verbose)
+			{
+				std::cerr << "INFO: machine-readablerevocation code means" <<
+					" that key is retired and no longer used" << std::endl;
+			}
+			break;
+		case TMCG_OPENPGP_REVCODE_UID_NO_LONGER_VALID:
+			std::cerr << "WARNING: machine-readable revocation code is" <<
+				" not admissible for keys and ignored" << std::endl;
+			revcode = TMCG_OPENPGP_REVCODE_NO_REASON_SPECIFIED;
+			break;
+		default:
+			std::cerr << "WARNING: machine-readable revocation code is" <<
+				" unknown" << std::endl;
+			break;
+	}
+
+	// compute a hash of pub and sub, respectively, and finally sign it
 	// RFC 4880 ERRATA:
 	// Primary key revocation signatures (type 0x20) hash only the key being
 	// revoked. Subkey revocation signature (type 0x28) hash first the primary
 	// key and then the subkey being revoked.
+	tmcg_openpgp_octets_t trailer_pub, pub_hashing, hash_pub, left_pub,
+		trailer_sub, sub_hashing, hash_sub, left_sub, issuer;
+	tmcg_openpgp_octets_t revsig_pub, revsig_sub;
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		FingerprintCompute(prv->pub->pub_hashing, issuer);
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		PacketSigPrepareRevocationSignature(TMCG_OPENPGP_SIGNATURE_KEY_REVOCATION,
+			hashalgo, csigtime, revcode, reason, issuer, trailer_pub);
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		KeyHash(prv->pub->pub_hashing, trailer_pub, hashalgo, hash_pub,
 			left_pub);
+	if (!sign_hash(hash_pub, trailer_pub, left_pub, whoami, peers.size(),
+		prv, hashalgo, revsig_pub, opt_verbose, NULL, dss, aiou, rbc))
+	{
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete prv;
+		exit(-1);
+	}
 	if (sub != NULL)
 	{
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			PacketSigPrepareRevocationSignature(TMCG_OPENPGP_SIGNATURE_SUBKEY_REVOCATION,
+				hashalgo, csigtime, revcode, reason, issuer, trailer_sub);
 		CallasDonnerhackeFinneyShawThayerRFC4880::
 			KeyHash(prv->pub->pub_hashing, sub->pub->sub_hashing, trailer_sub,
 				hashalgo, hash_sub, left_sub);
-	}
-
-	// sign the hashes
-	tmcg_openpgp_octets_t revsig_pub, revsig_sub;
-	tmcg_openpgp_byte_t buffer[1024];
-	gcry_mpi_t r, s, h;
-	mpz_t dsa_m, dsa_r, dsa_s;
-	size_t buflen = 0;
-	gcry_error_t ret;
-	memset(buffer, 0, sizeof(buffer));
-	for (size_t i = 0; ((i < hash_pub.size()) && (i < sizeof(buffer)));
-		i++, buflen++)
-			buffer[i] = hash_pub[i];
-	r = gcry_mpi_new(2048);
-	s = gcry_mpi_new(2048);
-	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
-	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-	if (ret)
-	{
-		std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed" <<
-			" for h" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
-			" failed for dsa_m" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	gcry_mpi_release(h);
-	std::stringstream err_log_sign;
-	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": dss.Sign() on" <<
-			" primary key" << std::endl;
-	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, prv->tdss_idx2dkg,
-		prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on" <<
-			" primary key failed" << std::endl;
-		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl <<
-			err_log_sign.str();
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	if (opt_verbose > 1)
-		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl <<
-			err_log_sign.str();
-	if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-			" failed for dsa_r" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-			" failed for dsa_s" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_pub,
-		left_pub, r, s, revsig_pub);
-	if (sub != NULL)
-	{
-		buflen = 0;
-		memset(buffer, 0, sizeof(buffer));
-		for (size_t i = 0; ((i < hash_sub.size()) && (i < sizeof(buffer)));
-			i++, buflen++)
-				buffer[i] = hash_sub[i];
-		ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-		if (ret)
+		if (!sign_hash(hash_sub, trailer_sub, left_sub, whoami, peers.size(),
+			prv, hashalgo, revsig_sub, opt_verbose, NULL, dss, aiou, rbc))
 		{
-			std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed" <<
-				" for h" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 			delete rbc, delete aiou, delete aiou2;
 			delete dss;
-			delete ring;
 			delete prv;
 			exit(-1);
 		}
-		if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
-		{
-			std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
-				" failed for dsa_m" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		gcry_mpi_release(h);
-		std::stringstream err_log_sign2;
-		if (opt_verbose)
-			std::cerr << "INFO: R_" << whoami << ": dss.Sign() on" <<
-				" first subkey" << std::endl;
-		if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s,
-			prv->tdss_idx2dkg, prv->tdss_dkg2idx, aiou, rbc, err_log_sign2))
-		{
-			std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on" <<
-				" first subkey failed" << std::endl;
-			std::cerr << "ERROR: R_" << whoami << ": log follows " <<
-				std::endl << err_log_sign2.str();
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (opt_verbose > 1)
-			std::cerr << "INFO: R_" << whoami << ": log follows " <<
-				std::endl << err_log_sign2.str();
-		if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
-		{
-			std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-				" failed for dsa_r" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
-		{
-			std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-				" failed for dsa_s" << std::endl;
-			gcry_mpi_release(r), gcry_mpi_release(s);
-			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
-			delete prv;
-			exit(-1);
-		}
-		CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_sub,
-			left_sub, r, s, revsig_sub);
 	}
-	gcry_mpi_release(r), gcry_mpi_release(s);
-	mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 
 	// at the end: deliver some more rounds for still waiting parties
 	time_t synctime = aiounicast::aio_timeout_long;
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": waiting approximately " <<
+		std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
 			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
 			std::endl;
 	rbc->Sync(synctime);
 
-	// export updated public key in OpenPGP armor format
-	std::stringstream pubfilename;
-	pubfilename << peers[whoami] << "_dkg-pub.asc";
-	std::string armor = "";
-	tmcg_openpgp_octets_t all;
-	all.insert(all.end(),
-		prv->pub->packet.begin(), prv->pub->packet.end());
-	all.insert(all.end(), revsig_pub.begin(), revsig_pub.end());
-	for (size_t k = 0; k < prv->pub->selfsigs.size(); k++)
-		all.insert(all.end(),
-			(prv->pub->selfsigs[k]->packet).begin(),
-			(prv->pub->selfsigs[k]->packet).end());
-	for (size_t k = 0; k < prv->pub->keyrevsigs.size(); k++)
-		all.insert(all.end(),
-			(prv->pub->keyrevsigs[k]->packet).begin(),
-			(prv->pub->keyrevsigs[k]->packet).end());
-	for (size_t j = 0; j < prv->pub->userids.size(); j++)
-	{
-		TMCG_OpenPGP_UserID *uid = prv->pub->userids[j];
-		if (uid->valid)
-		{
-			all.insert(all.end(),
-				(uid->packet).begin(), (uid->packet).end());
-			for (size_t k = 0; k < uid->selfsigs.size(); k++)
-				all.insert(all.end(),
-					(uid->selfsigs[k]->packet).begin(),
-					(uid->selfsigs[k]->packet).end());
-			for (size_t k = 0; k < uid->revsigs.size(); k++)
-				all.insert(all.end(),
-					(uid->selfsigs[k]->packet).begin(),
-					(uid->selfsigs[k]->packet).end());
-		}
-	}
-	for (size_t j = 0; j < prv->pub->userattributes.size(); j++)
-	{
-		TMCG_OpenPGP_UserAttribute *uat = prv->pub->userattributes[j];
-		if (uat->valid)
-		{
-			all.insert(all.end(),
-				(uat->packet).begin(), (uat->packet).end());
-			for (size_t k = 0; k < uat->selfsigs.size(); k++)
-				all.insert(all.end(),
-					(uat->selfsigs[k]->packet).begin(),
-					(uat->selfsigs[k]->packet).end());
-			for (size_t k = 0; k < uat->revsigs.size(); k++)
-				all.insert(all.end(),
-					(uat->selfsigs[k]->packet).begin(),
-					(uat->selfsigs[k]->packet).end());
-		}
-	}
-	if (sub != NULL)
-	{
-		TMCG_OpenPGP_Subkey *psub = sub->pub;
-		all.insert(all.end(),
-			psub->packet.begin(), psub->packet.end());
-		all.insert(all.end(), revsig_sub.begin(), revsig_sub.end());
-		for (size_t k = 0; k < psub->selfsigs.size(); k++)
-			all.insert(all.end(),
-				(psub->selfsigs[k]->packet).begin(),
-				(psub->selfsigs[k]->packet).end());
-		for (size_t k = 0; k < psub->bindsigs.size(); k++)
-			all.insert(all.end(),
-				(psub->bindsigs[k]->packet).begin(),
-				(psub->bindsigs[k]->packet).end());
-		for (size_t k = 0; k < psub->pbindsigs.size(); k++)
-			all.insert(all.end(),
-				(psub->pbindsigs[k]->packet).begin(),
-				(psub->pbindsigs[k]->packet).end());
-		for (size_t k = 0; k < psub->keyrevsigs.size(); k++)
-			all.insert(all.end(),
-				(psub->keyrevsigs[k]->packet).begin(),
-				(psub->keyrevsigs[k]->packet).end());
-	}
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
-	if (opt_verbose > 1)
-		std::cout << armor << std::endl;
-	std::ofstream pubofs((pubfilename.str()).c_str(), std::ofstream::out);
-	if (!pubofs.good())
-	{
-		std::cerr << "ERROR: R_" << whoami << ": opening public key file" <<
-			" failed" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	pubofs << armor;
-	if (!pubofs.good())
-	{
-		std::cerr << "ERROR: R_" << whoami << ": writing public key file" <<
-			" failed" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	pubofs.close();
-
 	// release RBC
 	delete rbc;
 	
@@ -647,7 +365,7 @@ void run_instance
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
 	{
-		std::cerr << "INFO: R_" << whoami << ": unicast channels";
+		std::cerr << "INFO: p_" << whoami << ": unicast channels";
 		aiou->PrintStatistics(std::cerr);
 		std::cerr << std::endl;
 	}
@@ -656,7 +374,7 @@ void run_instance
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
 	{
-		std::cerr << "INFO: R_" << whoami << ": broadcast channel";
+		std::cerr << "INFO: p_" << whoami << ": broadcast channel";
 		aiou2->PrintStatistics(std::cerr);
 		std::cerr << std::endl;
 	}
@@ -666,7 +384,66 @@ void run_instance
 
 	// release
 	delete dss;
-	delete ring;
+
+	// convert and append the created signature packets (revsig_pub, revsig_sub)
+	// to existing OpenPGP structures of this key
+	TMCG_OpenPGP_Signature *si_pub = NULL, *si_sub = NULL;
+	bool parse_ok_pub = CallasDonnerhackeFinneyShawThayerRFC4880::
+		SignatureParse(revsig_pub, opt_verbose, si_pub);
+	if (!parse_ok_pub)
+	{
+		std::cerr << "ERROR: cannot use the created signature" << std::endl;
+		delete prv;
+		exit(-1);
+	}
+	if (opt_verbose)
+		si_pub->PrintInfo();
+	prv->pub->keyrevsigs.push_back(si_pub); // append to private/public key
+	if (sub != NULL)
+	{
+		bool parse_ok_sub = CallasDonnerhackeFinneyShawThayerRFC4880::
+			SignatureParse(revsig_sub, opt_verbose, si_sub);
+		if (!parse_ok_sub)
+		{
+			std::cerr << "ERROR: cannot use the created signature" << std::endl;
+			delete prv;
+			exit(-1);
+		}
+		if (opt_verbose)
+			si_sub->PrintInfo();
+		sub->pub->keyrevsigs.push_back(si_sub); // append to private/public key
+	}
+
+	// export and write updated private key in OpenPGP armor format
+	std::stringstream secfilename;
+	secfilename << peers[whoami] << "_dkg-sec.asc";
+	tmcg_openpgp_octets_t sec;
+	prv->Export(sec);
+	if (!write_key_file(secfilename.str(),
+		TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, sec))
+	{
+		delete prv;
+		exit(-1);
+	}
+
+	// export and write updated public key in OpenPGP armor format
+	std::stringstream pubfilename;
+	pubfilename << peers[whoami] << "_dkg-pub.asc";
+	tmcg_openpgp_octets_t pub;
+	prv->RelinkPublicSubkeys(); // relink the contained subkeys
+	prv->pub->Export(pub);
+	prv->RelinkPrivateSubkeys(); // undo the relinking
+	std::string armor;
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, pub, armor);
+	std::cout << armor << std::endl;
+	if (!write_key_file(pubfilename.str(), armor))
+	{
+		delete prv;
+		exit(-1);
+	}
+
+	// release
 	delete prv;
 }
 
@@ -675,7 +452,8 @@ char *gnunet_opt_hostname = NULL;
 char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 char *gnunet_opt_k = NULL;
-unsigned int gnunet_opt_reason = 0;
+char *gnunet_opt_R = NULL;
+unsigned int gnunet_opt_r = 0;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
 unsigned int gnunet_opt_W = opt_W;
@@ -691,17 +469,17 @@ void fork_instance
 	{
 		if (pid[whoami] == 0)
 		{
-			/* BEGIN child code: participant R_i */
+			/* BEGIN child code: participant p_i */
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
-			run_instance(whoami, sigtime, gnunet_opt_reason, gnunet_opt_xtests);
+			run_instance(whoami, sigtime, gnunet_opt_r, gnunet_opt_xtests);
 #else
 			run_instance(whoami, sigtime, opt_r, 0);
 #endif
 			if (opt_verbose)
-				std::cerr << "INFO: R_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: p_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
-			/* END child code: participant R_i */
+			/* END child code: participant p_i */
 		}
 		else
 		{
@@ -753,7 +531,13 @@ int main
 			"reason",
 			"INTEGER",
 			"reason for revocation (OpenPGP machine-readable code)",
-			&gnunet_opt_reason
+			&gnunet_opt_r
+		),
+		GNUNET_GETOPT_option_string('R',
+			"Reason",
+			"STRING",
+			"reason for revocation (human-readable form)",
+			&gnunet_opt_R
 		),
 		GNUNET_GETOPT_option_version(version),
 		GNUNET_GETOPT_option_flag('V',
@@ -763,13 +547,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of revocation protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -801,6 +585,11 @@ int main
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
 	if (gnunet_opt_k != NULL)
 		opt_k = gnunet_opt_k;
+	if (gnunet_opt_R != NULL)
+	{
+		reason = gnunet_opt_R; // get reason string from GNUnet options
+		opt_R = gnunet_opt_R;
+	}
 	if (gnunet_opt_W != opt_W)
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 #endif
@@ -815,10 +604,11 @@ int main
 			(arg.find("-W") == 0) || (arg.find("-L") == 0) ||
 			(arg.find("-l") == 0) || (arg.find("-x") == 0) ||
 			(arg.find("-P") == 0) || (arg.find("-H") == 0) ||
-			(arg.find("-k") == 0))
+			(arg.find("-k") == 0) || (arg.find("-R") == 0))
 		{
 			size_t idx = ++i;
-			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) && (opt_hostname == NULL))
+			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_hostname == NULL))
 			{
 				hostname = argv[i+1];
 				opt_hostname = (char*)hostname.c_str();
@@ -829,44 +619,72 @@ int main
 				kfilename = argv[i+1];
 				opt_k = (char*)kfilename.c_str();
 			}
-			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) && (opt_passwords == NULL))
+			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_passwords == NULL))
 			{
 				passwords = argv[i+1];
 				opt_passwords = (char*)passwords.c_str();
 			}
-			if ((arg.find("-r") == 0) && (idx < (size_t)(argc - 1)) && (opt_r == 0))
-				opt_r = strtoul(argv[i+1], NULL, 10);
-			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) && (port.length() == 0))
+			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) &&
+				(port.length() == 0))
+			{
 				port = argv[i+1];
-			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) && (opt_W == 5))
+			}
+			if ((arg.find("-R") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_R == NULL))
+			{
+				reason = argv[i+1];
+				opt_R = (char*)reason.c_str();
+			}
+			if ((arg.find("-r") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_r == 0))
+			{
+				opt_r = strtoul(argv[i+1], NULL, 10);
+			}
+			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_W == 5))
+			{
 				opt_W = strtoul(argv[i+1], NULL, 10);
+			}
 			continue;
 		}
-		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-V") == 0))
+		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) ||
+				 (arg.find("-h") == 0) || (arg.find("-V") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
 #ifndef GNUNET
 				std::cout << usage << std::endl;
 				std::cout << about << std::endl;
-				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
+				std::cout << "Arguments mandatory for long options are also" <<
+					" mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
-				std::cout << "  -H STRING      hostname (e.g. onion address) of this peer within PEERS" << std::endl;
+				std::cout << "  -H STRING      hostname (e.g. onion address)" <<
+					" of this peer within PEERS" << std::endl;
 				std::cout << "  -k FILENAME    use keyring FILENAME" <<
 					" containing external revocation keys" << std::endl;
-				std::cout << "  -p INTEGER     start port for built-in TCP/IP message exchange service" << std::endl;
-				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
-				std::cout << "  -r INTEGER     reason for revocation (OpenPGP machine-readable code)" << std::endl;
-				std::cout << "  -v, --version  print the version number" << std::endl;
-				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point messages in minutes" << std::endl;
+				std::cout << "  -p INTEGER     start port for built-in" <<
+					" TCP/IP message exchange service" << std::endl;
+				std::cout << "  -P STRING      exchanged passwords to" <<
+					" protect private and broadcast channels" << std::endl;
+				std::cout << "  -r INTEGER     reason for revocation" <<
+					" (OpenPGP machine-readable code)" << std::endl;
+				std::cout << "  -R STRING      reason for revocation" <<
+					" (human-readable form)" << std::endl;
+				std::cout << "  -v, --version  print the version number" <<
+					std::endl;
+				std::cout << "  -V, --verbose  turn on verbose output" <<
+					std::endl;
+				std::cout << "  -W INTEGER     timeout for point-to-point" <<
+					" messages in minutes" << std::endl;
 #endif
 				return 0; // not continue
 			}
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
 			{
 #ifndef GNUNET
-				std::cout << "dkg-revoke v" << version << " without GNUNET support" << std::endl;
+				std::cout << "dkg-revoke v" << version <<
+					" without GNUNET support" << std::endl;
 #endif
 				return 0; // not continue
 			}
@@ -886,7 +704,8 @@ int main
 		}
 		else
 		{
-			std::cerr << "ERROR: peer identity \"" << arg << "\" too long" << std::endl;
+			std::cerr << "ERROR: peer identity \"" << arg << "\" too long" <<
+				std::endl;
 			return -1;
 		}
 	}
@@ -895,24 +714,29 @@ int main
 	peers.push_back("Test3");
 	peers.push_back("Test4");
 	opt_r = 3;
+	reason = "PGP is dead";
+	opt_R = (char*)reason.c_str();
 	opt_verbose = 2;
 #endif
 
 	// check command line arguments
 	if ((opt_hostname != NULL) && (opt_passwords == NULL))
 	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		std::cerr << "ERROR: option \"-P\" is necessary due to insecure" <<
+			" network" << std::endl;
 		return -1;
 	}
 	if (peers.size() < 1)
 	{
-		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
+		std::cerr << "ERROR: no peers given as argument; usage: " <<
+			usage << std::endl;
 		return -1;
 	}
 
 	// canonicalize peer list
 	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
+	std::vector<std::string>::iterator it =
+		std::unique(peers.begin(), peers.end());
 	peers.resize(std::distance(peers.begin(), it));
 	if ((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N))
 	{
@@ -943,15 +767,16 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
+			std::endl;
 	
 	// initialize return code
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP service requested
+	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if (opt_hostname != NULL)
 	{
 		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // get start port from options
+			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
 		if ((opt_p < 1) || (opt_p > 65535))
 		{
 			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
@@ -1003,7 +828,13 @@ int main
 			"reason",
 			"INTEGER",
 			"reason for revocation (OpenPGP machine-readable code)",
-			&gnunet_opt_reason
+			&gnunet_opt_r
+		),
+		GNUNET_GETOPT_option_string('R',
+			"Reason",
+			"STRING",
+			"reason for revocation (human-readable form)",
+			&gnunet_opt_R
 		),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
@@ -1012,13 +843,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of revocation protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -1030,17 +861,20 @@ int main
 		),
 		GNUNET_GETOPT_OPTION_END
 	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run, argv[0]);
+	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
+		argv[0]);
 //	GNUNET_free((void *) argv);
 	if (ret == GNUNET_OK)
 		return 0;
 	else
 		return -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
+	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
+		" of this program" << std::endl;
 #endif
 
-	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() <<
+		" participants" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -1072,7 +906,10 @@ int main
 		{
 			std::cerr << "ERROR: protocol instance ";
 			if (WIFSIGNALED(wstatus))
-				std::cerr << pid[i] << " terminated by signal " << WTERMSIG(wstatus) << std::endl;
+			{
+				std::cerr << pid[i] << " terminated by signal " <<
+					WTERMSIG(wstatus) << std::endl;
+			}
 			if (WCOREDUMP(wstatus))
 				std::cerr << pid[i] << " dumped core" << std::endl;
 			ret = -1;
@@ -1080,7 +917,11 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+			{
+				std::cerr << "INFO: protocol instance " << pid[i] <<
+					" terminated with exit status " << WEXITSTATUS(wstatus) <<
+					std::endl;
+			}
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
@@ -1088,8 +929,11 @@ int main
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
 				perror("ERROR: dkg-revoke (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
+			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
+				(close(broadcast_pipefd[i][j][1]) < 0))
+			{
 				perror("ERROR: dkg-revoke (close)");
+			}
 		}
 	}
 	
