@@ -24,11 +24,16 @@
 #endif
 #ifdef DKGPG_TESTSUITE
 	#undef GNUNET
+#else
+#ifdef DKGPG_TESTSUITE_Y
+	#undef GNUNET
+#endif
 #endif
 
 // copy infos from DKGPG package before overwritten by GNUnet headers
 static const char *version = PACKAGE_VERSION " (" PACKAGE_NAME ")";
 static const char *about = PACKAGE_STRING " " PACKAGE_URL;
+static const char *protocol = "DKGPG-sign-1.0";
 
 #include <sstream>
 #include <fstream>
@@ -57,40 +62,49 @@ pid_t 						pid[DKGPG_MAX_N];
 std::vector<std::string>	peers;
 bool						instance_forked = false;
 
-std::string					passphrase, ifilename, ofilename, kfilename;
-std::string					passwords, hostname, port, URI;
+tmcg_openpgp_secure_string_t	passphrase;
+std::string						ifilename, ofilename, kfilename;
+std::string						passwords, hostname, port, URI, yfilename;
 
-int 						opt_verbose = 0;
-char						*opt_ifilename = NULL;
-char						*opt_ofilename = NULL;
-char						*opt_passwords = NULL;
-char						*opt_hostname = NULL;
-char						*opt_URI = NULL;
-char						*opt_k = NULL;
-unsigned long int			opt_e = 0, opt_p = 55000, opt_W = 5;
-bool						opt_t = false;
+int 							opt_verbose = 0;
+char							*opt_ifilename = NULL;
+char							*opt_ofilename = NULL;
+char							*opt_passwords = NULL;
+char							*opt_hostname = NULL;
+char							*opt_URI = NULL;
+char							*opt_k = NULL;
+char							*opt_y = NULL;
+unsigned long int				opt_e = 0, opt_p = 55000, opt_W = 5;
+bool							opt_t = false, opt_E = false, opt_C = false;
 
 void run_instance
 	(size_t whoami, const time_t sigtime, const time_t sigexptime,
 	 const size_t num_xtests)
 {
 	// read the key file
-	std::string armored_seckey, thispeer = peers[whoami];
-	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	std::string armored_seckey, pkfname;
+	if (opt_y == NULL)
+		pkfname = peers[whoami] + "_dkg-sec.asc";
+	else
+		pkfname = opt_y;
+	if (opt_verbose > 1)
+		std::cerr << "INFO: private key expected in file \"" << pkfname <<
+			"\"" << std::endl;
+	if (!check_strict_permissions(pkfname))
 	{
 		std::cerr << "WARNING: weak permissions of private key file" <<
 			" detected" << std::endl;
-		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+		if (!set_strict_permissions(pkfname))
 			exit(-1);
 	}
-	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
+	if (!read_key_file(pkfname, armored_seckey))
 		exit(-1);
 
 	// read the keyring
 	std::string armored_pubring;
 	if (opt_k)
 	{
-		if (!read_key_file(kfilename, armored_pubring))
+		if (!read_key_file(opt_k, armored_pubring))
 			exit(-1);
 	}
 
@@ -117,13 +131,17 @@ void run_instance
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		if (!get_passphrase("Enter passphrase to unlock private key",
+#ifdef DKGPG_TESTSUITE_Y
+		passphrase = "TestY";
+#else
+		if (!get_passphrase("Enter passphrase to unlock private key", opt_E,
 			passphrase))
 		{
 			std::cerr << "ERROR: cannot read passphrase" << std::endl;
 			delete ring;
 			exit(-1);
 		}
+#endif
 #endif
 		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
 			PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
@@ -141,373 +159,254 @@ void run_instance
 		delete ring;
 		exit(-1);
 	}
-	if (!prv->pub->valid || prv->weak(opt_verbose))
+	delete ring;
+	if (!prv->pub->valid || ((opt_y == NULL) && prv->Weak(opt_verbose)))
 	{
 		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
-	if (prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7)
+	if ((prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
-		delete ring;
 		delete prv;
 		exit(-1);
 	}
 
-	// create an instance of tDSS by stored parameters from private key
+	// initialize signature scheme
 	CanettiGennaroJareckiKrawczykRabinDSS *dss = NULL;
-	if (!init_tDSS(prv, opt_verbose, dss))
+	aiounicast_select *aiou = NULL, *aiou2 = NULL;
+	CachinKursawePetzoldShoupRBC *rbc = NULL;
+	size_t T_RBC = 0;
+	time_t csigtime = 0;
+	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
+	if (opt_y == NULL)
 	{
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	// create one-to-one mapping based on the stored canonicalized peer list
-	if (!prv->tDSS_CreateMapping(peers, opt_verbose))
-	{
-		std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" << std::endl;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-
-	// create communication handles between all players
-	std::vector<int> uP_in, uP_out, bP_in, bP_out;
-	std::vector<std::string> uP_key, bP_key;
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		std::stringstream key;
-		if (opt_passwords != NULL)
+		// create an instance of tDSS by stored parameters from private key
+		if (!init_tDSS(prv, opt_verbose, dss))
 		{
-			std::string pwd;
-			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
-			{
-				std::cerr << "ERROR: S_" << whoami << ": " << "cannot read" <<
-					" password for protecting channel to S_" << i << std::endl;
-				delete dss;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
-			key << pwd;
-			if (((i + 1) < peers.size()) &&
-				!TMCG_ParseHelper::nx(passwords, '/'))
-			{
-				std::cerr << "ERROR: S_" << whoami << ": " << "cannot skip" <<
-					" to next password for protecting channel to S_" <<
-					(i + 1) << std::endl;
-				delete dss;
-				delete ring;
-				delete prv;
-				exit(-1);
-			}
+			delete dss;
+			delete prv;
+			exit(-1);
 		}
+		// create one-to-one mapping based on the stored canonicalized peer list
+		if (!prv->tDSS_CreateMapping(peers, opt_verbose))
+		{
+			std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" <<
+				std::endl;
+			delete dss;
+			delete prv;
+			exit(-1);
+		}
+		// create communication handles between all players
+		std::vector<int> uP_in, uP_out, bP_in, bP_out;
+		std::vector<std::string> uP_key, bP_key;
+		for (size_t i = 0; i < peers.size(); i++)
+		{
+			std::stringstream key;
+			if (opt_passwords != NULL)
+			{
+				std::string pwd;
+				if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
+				{
+					std::cerr << "ERROR: p_" << whoami << ": " <<
+						"cannot read password for protecting channel to p_" <<
+						i << std::endl;
+					delete dss;
+					delete prv;
+					exit(-1);
+				}
+				key << pwd;
+				if (((i + 1) < peers.size()) &&
+					!TMCG_ParseHelper::nx(passwords, '/'))
+				{
+					std::cerr << "ERROR: p_" << whoami << ": " << "cannot" <<
+						" skip to next password for protecting channel to p_" <<
+						(i + 1) << std::endl;
+					delete dss;
+					delete prv;
+					exit(-1);
+				}
+			}
+			else
+			{
+				// simple key -- we assume that GNUnet provides secure channels
+				key << "dkg-sign::p_" << (i + whoami);
+			}
+			uP_in.push_back(pipefd[i][whoami][0]);
+			uP_out.push_back(pipefd[whoami][i][1]);
+			uP_key.push_back(key.str());
+			bP_in.push_back(broadcast_pipefd[i][whoami][0]);
+			bP_out.push_back(broadcast_pipefd[whoami][i][1]);
+			bP_key.push_back(key.str());
+		}
+		// create asynchronous authenticated unicast channels
+		aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out,
+			uP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+		// create asynchronous authenticated unicast channels for broadcast
+		aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out,
+			bP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+		// create an instance of a reliable broadcast protocol (RBC)
+		std::string myID = "dkg-sign|" + std::string(protocol) + "|";
+		for (size_t i = 0; i < peers.size(); i++)
+			myID += peers[i] + "|";
+		if (opt_verbose)
+			std::cerr << "RBC: myID = " << myID << std::endl;
+		// assume maximum asynchronous t-resilience for RBC
+		T_RBC = (peers.size() - 1) / 3;
+		rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami,
+				aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+		rbc->setID(myID);
+		// perform a simple exchange test with debug output
+		xtest(num_xtests, whoami, peers.size(), rbc);
+		// participants must agree on a common signature creation time (OpenPGP)
+		csigtime = agree_time(sigtime, whoami, peers.size(), opt_verbose, rbc);
+		// select hash algorithm for OpenPGP based on |q| (size in bit)
+		if (mpz_sizeinbase(dss->q, 2L) == 256)
+			hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
+		else if (mpz_sizeinbase(dss->q, 2L) == 384)
+			hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
+		else if (mpz_sizeinbase(dss->q, 2L) == 512)
+			hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 		else
 		{
-			// use simple key -- we assume that GNUnet provides secure channels
-			key << "dkg-sign::S_" << (i + whoami);
-		}
-		uP_in.push_back(pipefd[i][whoami][0]);
-		uP_out.push_back(pipefd[whoami][i][1]);
-		uP_key.push_back(key.str());
-		bP_in.push_back(broadcast_pipefd[i][whoami][0]);
-		bP_out.push_back(broadcast_pipefd[whoami][i][1]);
-		bP_key.push_back(key.str());
-	}
-
-	// create asynchronous authenticated unicast channels
-	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami,
-		uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin,
-		(opt_W * 60));
-
-	// create asynchronous authenticated unicast channels for broadcast protocol
-	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami,
-		bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin,
-		(opt_W * 60));
-			
-	// create an instance of a reliable broadcast protocol (RBC)
-	std::string myID = "dkg-sign|";
-	for (size_t i = 0; i < peers.size(); i++)
-		myID += peers[i] + "|";
-	// assume maximum asynchronous t-resilience for RBC
-	size_t T_RBC = (peers.size() - 1) / 3;
-	CachinKursawePetzoldShoupRBC *rbc =
-		new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2,
-			aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
-	rbc->setID(myID);
-
-	// perform a simple exchange test with debug output
-	for (size_t i = 0; i < num_xtests; i++)
-	{
-		mpz_t xtest;
-		mpz_init_set_ui(xtest, i);
-		std::cerr << "INFO: S_" << whoami << ": xtest = " << xtest << " <-> ";
-		rbc->Broadcast(xtest);
-		for (size_t ii = 0; ii < peers.size(); ii++)
-		{
-			if (!rbc->DeliverFrom(xtest, ii))
-				std::cerr << "<X> ";
-			else
-				std::cerr << xtest << " ";
-		}
-		std::cerr << std::endl;
-		mpz_clear(xtest);
-	}
-
-	// participants must agree on a common signature creation time (OpenPGP)
-	if (opt_verbose)
-		std::cerr << "INFO: agree on a signature creation time for OpenPGP" <<
-			std::endl;
-	time_t csigtime = 0;
-	std::vector<time_t> tvs;
-	mpz_t mtv;
-	mpz_init_set_ui(mtv, sigtime);
-	rbc->Broadcast(mtv);
-	tvs.push_back(sigtime);
-	for (size_t i = 0; i < peers.size(); i++)
-	{
-		if (i != whoami)
-		{
-			if (rbc->DeliverFrom(mtv, i))
-			{
-				time_t utv;
-				utv = (time_t)mpz_get_ui(mtv);
-				tvs.push_back(utv);
-			}
-			else
-			{
-				std::cerr << "WARNING: S_" << whoami << ": no signature" <<
-					" creation time stamp received from S_" << i << std::endl;
-			}
+			std::cerr << "ERROR: p_" << whoami << ": selecting hash" <<
+				" algorithm failed for |q| = " << mpz_sizeinbase(dss->q, 2L) <<
+				std::endl;
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete prv;
+			exit(-1);
 		}
 	}
-	mpz_clear(mtv);
-	std::sort(tvs.begin(), tvs.end());
-	if (tvs.size() < (peers.size() - T_RBC))
-	{
-		std::cerr << "ERROR: S_" << whoami << ": not enough timestamps" <<
-			" received" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	// use a median value as some kind of gentle agreement
-	csigtime = tvs[tvs.size()/2];
-	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": canonicalized signature" <<
-			" creation time = " << csigtime << std::endl;
-
-	// select hash algorithm for OpenPGP based on |q| (size in bit)
-	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
-	if (mpz_sizeinbase(dss->q, 2L) == 256)
-		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-	else if (mpz_sizeinbase(dss->q, 2L) == 384)
-		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-	else if (mpz_sizeinbase(dss->q, 2L) == 512)
-		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 	else
 	{
-		std::cerr << "ERROR: S_" << whoami << ": selecting hash algorithm" <<
-			" failed for |q| = " << mpz_sizeinbase(dss->q, 2L) << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
+		csigtime = sigtime;
+		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // fixed hash algo SHA2-512
 	}
 
 	// compute the hash of the input file
 	if (opt_verbose)
 		std::cerr << "INFO: hashing the input file \"" << opt_ifilename <<
 			"\"" << std::endl;
-	tmcg_openpgp_octets_t trailer, hash, left;
-	if (opt_t)
+	tmcg_openpgp_octets_t trailer, hash, left, issuer;
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		FingerprintCompute(prv->pub->pub_hashing, issuer);	
+	if (opt_t || opt_C)
 	{
-		CallasDonnerhackeFinneyShawThayerRFC4880::
-			PacketSigPrepareDetachedSignature(
-				TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT,
-				hashalgo, csigtime, sigexptime, URI, prv->pub->id, trailer);
+		if (opt_y == NULL)
+		{
+			CallasDonnerhackeFinneyShawThayerRFC4880::
+				PacketSigPrepareDetachedSignature(
+					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT,
+					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+		}
+		else
+		{
+			CallasDonnerhackeFinneyShawThayerRFC4880::
+				PacketSigPrepareDetachedSignature(
+					TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT, prv->pkalgo,
+					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+		}
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			TextDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": TextDocumentHash()" <<
+			std::cerr << "ERROR: p_" << whoami << ": TextDocumentHash()" <<
 				" failed; cannot process input file \"" << opt_ifilename <<
 				"\"" << std::endl;
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
+			if (opt_y == NULL)
+			{
+				delete rbc, delete aiou, delete aiou2;
+				delete dss;
+			}
 			delete prv;
 			exit(-1);
 		}
 	}
 	else
 	{
-		CallasDonnerhackeFinneyShawThayerRFC4880::
-			PacketSigPrepareDetachedSignature(
-				TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT,
-				hashalgo, csigtime, sigexptime, URI, prv->pub->id, trailer);
+		if (opt_y == NULL)
+		{
+			CallasDonnerhackeFinneyShawThayerRFC4880::
+				PacketSigPrepareDetachedSignature(
+					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT,
+					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+		}
+		else
+		{
+			CallasDonnerhackeFinneyShawThayerRFC4880::
+				PacketSigPrepareDetachedSignature(
+					TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT, prv->pkalgo,
+					hashalgo, csigtime, sigexptime, URI, issuer, trailer);
+		}
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": BinaryDocumentHash()" <<
+			std::cerr << "ERROR: p_" << whoami << ": BinaryDocumentHash()" <<
 				" failed; cannot process input file \"" << opt_ifilename <<
 				"\"" << std::endl;
-			delete rbc, delete aiou, delete aiou2;
-			delete dss;
-			delete ring;
+			if (opt_y == NULL)
+			{
+				delete rbc, delete aiou, delete aiou2;
+				delete dss;
+			}
 			delete prv;
 			exit(-1);
 		}
 	}
 
-	// sign the hash
-	tmcg_openpgp_byte_t buffer[1024];
-	gcry_mpi_t r, s, h;
-	mpz_t dsa_m, dsa_r, dsa_s;
-	size_t buflen = 0;
-	gcry_error_t ret;
-	memset(buffer, 0, sizeof(buffer));
-	if (opt_verbose)
-		std::cerr << std::hex << "INFO: hash = ";
-	for (size_t i = 0; i < hash.size(); i++, buflen++)
-	{
-		if (i < sizeof(buffer))
-			buffer[i] = hash[i];
-		if (opt_verbose)
-			std::cerr << (int)hash[i] << " ";
-	}
-	if (opt_verbose)
-		std::cerr << std::dec << std::endl;
-	r = gcry_mpi_new(2048);
-	s = gcry_mpi_new(2048);
-	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
-	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-	if (ret)
-	{
-		std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan() failed" <<
-			" for h" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	if (opt_verbose > 1)
-		std::cerr << "INFO: S_" << whoami << ": h = " << h << std::endl;
-	if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
-	{
-		std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
-			" failed for dsa_m" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	gcry_mpi_release(h);
-	std::stringstream err_log_sign;
-	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": dss.Sign()" << std::endl;
-	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, prv->tdss_idx2dkg,
-		prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
-	{
-		std::cerr << "ERROR: S_" << whoami << ": " << "tDSS Sign() failed" <<
-			std::endl;
-		std::cerr << "ERROR: S_" << whoami << ": log follows " << std::endl <<
-			err_log_sign.str();
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	if (opt_verbose > 1)
-		std::cerr << "INFO: S_" << whoami << ": log follows " << std::endl <<
-			err_log_sign.str(); 
-	if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
-	{
-		std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-			" failed for dsa_r" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
-	if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
-	{
-		std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
-			" failed for dsa_s" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete rbc, delete aiou, delete aiou2;
-		delete dss;
-		delete ring;
-		delete prv;
-		exit(-1);
-	}
+	// sign the hash value
 	tmcg_openpgp_octets_t sig;
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		PacketSigEncode(trailer, left, r, s, sig);
-	gcry_mpi_release(r), gcry_mpi_release(s);
-	mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-
-	// at the end: deliver some more rounds for still waiting parties
-	time_t synctime = aiounicast::aio_timeout_long;
-	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": waiting approximately " <<
-			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
-			std::endl;
-	rbc->Sync(synctime);
-
-	// release RBC
-	delete rbc;
-	
-	// release handles (unicast channel)
-	uP_in.clear(), uP_out.clear(), uP_key.clear();
-	if (opt_verbose)
+	if (!sign_hash(hash, trailer, left, whoami, peers.size(), prv, hashalgo,
+		sig, opt_verbose, opt_y, dss, aiou, rbc))
 	{
-		std::cerr << "INFO: S_" << whoami << ": unicast channels";
-		aiou->PrintStatistics(std::cerr);
-		std::cerr << std::endl;
+		if (opt_y == NULL)
+		{
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+		}
+		delete prv;
+		exit(-1);
 	}
 
-	// release handles (broadcast channel)
-	bP_in.clear(), bP_out.clear(), bP_key.clear();
-	if (opt_verbose)
+	// release allocated ressources
+	if (opt_y == NULL)
 	{
-		std::cerr << "INFO: S_" << whoami << ": broadcast channel";
-		aiou2->PrintStatistics(std::cerr);
-		std::cerr << std::endl;
+		// at the end: deliver some more rounds for still waiting parties
+		time_t synctime = aiounicast::aio_timeout_long;
+		if (opt_verbose)
+			std::cerr << "INFO: p_" << whoami << ": waiting approximately " <<
+				(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
+				std::endl;
+		rbc->Sync(synctime);
+		// release RBC
+		delete rbc;
+		// release handles (unicast channel)
+		if (opt_verbose)
+		{
+			std::cerr << "INFO: p_" << whoami << ": unicast channels";
+			aiou->PrintStatistics(std::cerr);
+			std::cerr << std::endl;
+		}
+		// release handles (broadcast channel)
+		if (opt_verbose)
+		{
+			std::cerr << "INFO: p_" << whoami << ": broadcast channel";
+			aiou2->PrintStatistics(std::cerr);
+			std::cerr << std::endl;
+		}
+		// release asynchronous unicast and broadcast
+		delete aiou, delete aiou2;
+		// release threshold signature scheme
+		delete dss;
 	}
-
-	// release asynchronous unicast and broadcast
-	delete aiou, delete aiou2;
-
-	// release
-	delete dss;
-	delete ring;
 	delete prv;
 
 	// output the result
 	std::string sigstr;
 	CallasDonnerhackeFinneyShawThayerRFC4880::
 		ArmorEncode(TMCG_OPENPGP_ARMOR_SIGNATURE, sig, sigstr);
-	if (opt_t)
+	if (opt_C)
 	{
 		std::string ct_head = "-----BEGIN PGP SIGNED MESSAGE-----\r\n";
 		std::string ct_hash; // construct corresponding Hash Armor Header
@@ -519,7 +418,7 @@ void run_instance
 		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
 			DashEscapeFile(opt_ifilename, ct_body))
 		{
-			std::cerr << "ERROR: S_" << whoami << ": DashEscapeFile()" <<
+			std::cerr << "ERROR: p_" << whoami << ": DashEscapeFile()" <<
 				" failed; cannot process input file \"" << opt_ifilename <<
 				"\"" << std::endl;
 			exit(-1);
@@ -543,12 +442,15 @@ char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 char *gnunet_opt_URI = NULL;
 char *gnunet_opt_k = NULL;
+char *gnunet_opt_y = NULL;
 unsigned int gnunet_opt_sigexptime = 0;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
 unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_verbose = 0;
-int gnunet_opt_text = 0;
+int gnunet_opt_t = 0;
+int gnunet_opt_E = 0;
+int gnunet_opt_C = 0;
 #endif
 
 void fork_instance
@@ -560,7 +462,7 @@ void fork_instance
 	{
 		if (pid[whoami] == 0)
 		{
-			/* BEGIN child code: participant S_i */
+			/* BEGIN child code: participant p_i */
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
 			run_instance(whoami, sigtime, gnunet_opt_sigexptime,
@@ -569,9 +471,9 @@ void fork_instance
 			run_instance(whoami, sigtime, opt_e, 0);
 #endif
 			if (opt_verbose)
-				std::cerr << "INFO: S_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: p_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
-			/* END child code: participant S_i */
+			/* END child code: participant p_i */
 		}
 		else
 		{
@@ -592,13 +494,23 @@ int main
 	char *cfg_fn = NULL;
 	static const struct GNUNET_GETOPT_CommandLineOption options[] = {
 		GNUNET_GETOPT_option_cfgfile(&cfg_fn),
+		GNUNET_GETOPT_option_flag('C',
+			"clear",
+			"apply cleartext signature framework (cf. RFC 4880)",
+			&gnunet_opt_C
+		),
 		GNUNET_GETOPT_option_help(about),
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
-			"TIME",
+			"INTEGER",
 			"expiration time of generated signature in seconds",
 			&gnunet_opt_sigexptime
 		),
+		GNUNET_GETOPT_option_flag('E',
+			"echo",
+			"enable terminal echo when reading passphrase",
+			&gnunet_opt_E
+		),
 		GNUNET_GETOPT_option_string('H',
 			"hostname",
 			"STRING",
@@ -639,8 +551,8 @@ int main
 		),
 		GNUNET_GETOPT_option_flag('t',
 			"text",
-			"create cleartext signature",
-			&gnunet_opt_text
+			"create canonical text document signature",
+			&gnunet_opt_t
 		),
 		GNUNET_GETOPT_option_string('U',
 			"URI",
@@ -656,16 +568,22 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of signing protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
+		GNUNET_GETOPT_option_string('y',
+			"yaot",
+			"FILNAME",
+			"yet another OpenPGP tool with private key in FILENAME",
+			&gnunet_opt_y
+		),
 		GNUNET_GETOPT_option_uint('x',
 			"x-tests",
 			NULL,
@@ -674,12 +592,12 @@ int main
 		),
 		GNUNET_GETOPT_OPTION_END
 	};
-	if (GNUNET_STRINGS_get_utf8_args(argc, argv, &argc, &argv) != GNUNET_OK)
-	{
-		std::cerr << "ERROR: GNUNET_STRINGS_get_utf8_args() failed" <<
-			std::endl;
-   		return -1;
-	}
+//	if (GNUNET_STRINGS_get_utf8_args(argc, argv, &argc, &argv) != GNUNET_OK)
+//	{
+//		std::cerr << "ERROR: GNUNET_STRINGS_get_utf8_args() failed" <<
+//			std::endl;
+//		return -1;
+//	}
 	if (GNUNET_GETOPT_run(usage, options, argc, argv) == GNUNET_SYSERR)
 	{
 		std::cerr << "ERROR: GNUNET_GETOPT_run() failed" << std::endl;
@@ -705,6 +623,8 @@ int main
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 	if (gnunet_opt_URI != NULL)
 		URI = gnunet_opt_URI; // get policy URI from GNUnet options
+	if (gnunet_opt_y != NULL)
+		opt_y = gnunet_opt_y;
 #endif
 
 	// create peer list from remaining arguments
@@ -718,7 +638,8 @@ int main
 			(arg.find("-i") == 0) || (arg.find("-o") == 0) || 
 		    (arg.find("-e") == 0) || (arg.find("-x") == 0) ||
 			(arg.find("-P") == 0) || (arg.find("-H") == 0) ||
-		    (arg.find("-U") == 0) || (arg.find("-k") == 0))
+		    (arg.find("-U") == 0) || (arg.find("-k") == 0) ||
+			(arg.find("-y") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) &&
@@ -772,11 +693,18 @@ int main
 			{
 				opt_W = strtoul(argv[i+1], NULL, 10);
 			}
+			if ((arg.find("-y") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_y == NULL))
+			{
+				yfilename = argv[i+1];
+				opt_y = (char*)yfilename.c_str();
+			}
 			continue;
 		}
 		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) ||
 			(arg.find("-h") == 0) || (arg.find("-V") == 0) ||
-			(arg.find("-t") == 0))
+			(arg.find("-t") == 0) || (arg.find("-E") == 0) ||
+			(arg.find("-C") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
@@ -785,9 +713,13 @@ int main
 				std::cout << about << std::endl;
 				std::cout << "Arguments mandatory for long options are also" <<
 					" mandatory for short options." << std::endl;
+				std::cout << "  -C, --clear    apply cleartext signature" <<
+					" framework (cf. RFC 4880)" << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
-				std::cout << "  -e TIME        expiration time of generated" <<
+				std::cout << "  -e INTEGER     expiration time of generated" <<
 					" signature in seconds" << std::endl;
+				std::cout << "  -E, --echo     enable terminal echo when" <<
+					" reading passphrase" << std::endl;
 				std::cout << "  -H STRING      hostname (e.g. onion address)" <<
 					" of this peer within PEERS" << std::endl;
 				std::cout << "  -i FILENAME    create signature from" <<
@@ -800,19 +732,25 @@ int main
 					" TCP/IP message exchange service" << std::endl;
 				std::cout << "  -P STRING      exchanged passwords to" <<
 					" protect private and broadcast channels" << std::endl;
-				std::cout << "  -t, --text     create cleartext signature" <<
-					std::endl;
+				std::cout << "  -t, --text     create canonical text" <<
+					" document signature" << std::endl;
 				std::cout << "  -U STRING      policy URI tied to generated" <<
 					" signatures" << std::endl;
 				std::cout << "  -v, --version  print the version number" <<
 					std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" <<
 					std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point" <<
+				std::cout << "  -W INTEGER     timeout for point-to-point" <<
 					" messages in minutes" << std::endl;
+				std::cout << "  -y FILENAME    yet another OpenPGP tool with" <<
+					" private key in FILENAME" << std::endl;
 #endif
 				return 0; // not continue
 			}
+			if ((arg.find("-C") == 0) || (arg.find("--clear") == 0))
+				opt_C = true;
+			if ((arg.find("-E") == 0) || (arg.find("--echo") == 0))
+				opt_E = true;
 			if ((arg.find("-t") == 0) || (arg.find("--text") == 0))
 				opt_t = true;
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
@@ -854,22 +792,34 @@ int main
 	opt_ofilename = (char*)ofilename.c_str();
 	URI = "https://savannah.nongnu.org/projects/dkgpg/";
 	opt_verbose = 2;
+#else
+#ifdef DKGPG_TESTSUITE_Y
+	yfilename = "TestY-sec.asc";
+	opt_y = (char*)yfilename.c_str();
+	ifilename = "TestY_output.asc";
+	opt_ifilename = (char*)ifilename.c_str();
+	ofilename = "TestY_output.sig";
+	opt_ofilename = (char*)ofilename.c_str();
+	opt_e = 4242;
+	URI = "https://savannah.nongnu.org/projects/dkgpg/";
+	opt_verbose = 2;
+#endif
 #endif
 
 	// check command line arguments
-	if ((opt_hostname != NULL) && (opt_passwords == NULL))
-	{
-		std::cerr << "ERROR: option \"-P\" required due to insecure network" <<
-			std::endl;
-		return -1;
-	}
 	if (opt_ifilename == NULL)
 	{
 		std::cerr << "ERROR: option \"-i\" required to specify an input file" <<
 			std::endl;
 		return -1;
 	}
-	if (peers.size() < 1)
+	if ((opt_hostname != NULL) && (opt_passwords == NULL) && (opt_y == NULL))
+	{
+		std::cerr << "ERROR: option \"-P\" required due to insecure network" <<
+			std::endl;
+		return -1;
+	}
+	if ((peers.size() < 1) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: no peers given as argument; usage: " <<
 			usage << std::endl;
@@ -881,12 +831,12 @@ int main
 	std::vector<std::string>::iterator it =
 		std::unique(peers.begin(), peers.end());
 	peers.resize(std::distance(peers.begin(), it));
-	if ((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N))
+	if (((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N)) && (opt_y == NULL))
 	{
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
 		return -1;
 	}
-	if (opt_verbose)
+	if (opt_verbose && (opt_y == NULL))
 	{
 		std::cerr << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
@@ -894,26 +844,17 @@ int main
 	}
 
 	// lock memory
+	bool force_secmem = false;
 	if (!lock_memory())
 	{
 		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
-			" for memory protection" << std::endl;
+			" for full memory protection" << std::endl;
 		// at least try to use libgcrypt's secure memory
-		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
-		{
-			std::cerr << "ERROR: libgcrypt version >= " <<
-				TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
-			return -1;
-		}
-		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
-		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
-		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
-		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
-		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		force_secmem = true;
 	}
 
 	// initialize LibTMCG
-	if (!init_libTMCG())
+	if (!init_libTMCG(force_secmem))
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
 		return -1;
@@ -925,7 +866,7 @@ int main
 	// initialize return code
 	int ret = 0;
 	// create underlying point-to-point channels, if built-in TCP/IP requested
-	if (opt_hostname != NULL)
+	if ((opt_hostname != NULL) && (opt_y == NULL))
 	{
 		if (port.length())
 			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
@@ -948,16 +889,32 @@ int main
 		tcpip_done();
 		return ret;
 	}
+	else if (opt_y != NULL)
+	{
+		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
+		run_instance(0, time(NULL), opt_e, 0);
+		return ret;
+	}
 
 	// start interactive variant with GNUnet or otherwise a local test
 #ifdef GNUNET
 	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
+		GNUNET_GETOPT_option_flag('C',
+			"clear",
+			"apply cleartext signature framework (cf. RFC 4880)",
+			&gnunet_opt_C
+		),
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
-			"TIME",
+			"INTEGER",
 			"expiration time of generated signature in seconds",
 			&gnunet_opt_sigexptime
 		),
+		GNUNET_GETOPT_option_flag('E',
+			"echo",
+			"enable terminal echo when reading passphrase",
+			&gnunet_opt_E
+		),
 		GNUNET_GETOPT_option_string('H',
 			"hostname",
 			"STRING",
@@ -1002,8 +959,8 @@ int main
 		),
 		GNUNET_GETOPT_option_flag('t',
 			"text",
-			"create cleartext signature",
-			&gnunet_opt_text
+			"create canonical text document signature",
+			&gnunet_opt_t
 		),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
@@ -1012,16 +969,22 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of signing protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
+		GNUNET_GETOPT_option_string('y',
+			"yaot",
+			"FILNAME",
+			"yet another OpenPGP tool with private key in FILENAME",
+			&gnunet_opt_y
+		),
 		GNUNET_GETOPT_option_uint('x',
 			"x-tests",
 			NULL,
@@ -1032,7 +995,7 @@ int main
 	};
 	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
 		argv[0]);
-	GNUNET_free((void *) argv);
+//	GNUNET_free((void *) argv);
 	if (ret == GNUNET_OK)
 		return 0;
 	else
@@ -1075,8 +1038,10 @@ int main
 		{
 			std::cerr << "ERROR: protocol instance ";
 			if (WIFSIGNALED(wstatus))
+			{
 				std::cerr << pid[i] << " terminated by signal " <<
 					WTERMSIG(wstatus) << std::endl;
+			}
 			if (WCOREDUMP(wstatus))
 				std::cerr << pid[i] << " dumped core" << std::endl;
 			ret = -1; // fatal error
@@ -1084,9 +1049,11 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
+			{
 				std::cerr << "INFO: protocol instance " << pid[i] <<
 					" terminated with exit status " << WEXITSTATUS(wstatus) <<
 					std::endl;
+			}
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
