@@ -114,7 +114,7 @@ void run_instance
 		// generate a non-shared DSA primary key
 		mpz_t dsa_y, dsa_x;
 		mpz_init(dsa_y), mpz_init(dsa_x);
-		tmcg_mpz_ssrandomm_cache(cache, cache_mod, &cache_avail, dsa_x, vtmf->q);
+		tmcg_mpz_ssrandomm_cache(cache, cache_mod, cache_avail, dsa_x, vtmf->q);
 		tmcg_mpz_spowm(dsa_y, vtmf->g, dsa_x, vtmf->p);
 		// extract parameters for OpenPGP key structures
 		std::string out, crcout, armor;
@@ -246,7 +246,7 @@ void run_instance
 		// generate a non-shared ElGamal subkey with same domain parameter set
 		mpz_t elg_y, elg_x;
 		mpz_init(elg_y), mpz_init(elg_x);
-		tmcg_mpz_ssrandomm_cache(cache, cache_mod, &cache_avail, elg_x, vtmf->q);
+		tmcg_mpz_ssrandomm_cache(cache, cache_mod, cache_avail, elg_x, vtmf->q);
 		tmcg_mpz_spowm(elg_y, vtmf->g, elg_x, vtmf->p);
 		// extract further parameters for OpenPGP key structures
 		y = gcry_mpi_new(2048);
@@ -472,10 +472,9 @@ void run_instance
 	for (size_t i = 0; i < peers.size(); i++)
 		myID += peers[i] + "|";
 	// include parameters in the ID of broadcast protocol to enforce equal set
-	myID += T;
-	myID += "|";
-	myID += S;
-	myID += "|";
+	std::stringstream myss;
+	myss << T << "|" << S << "|";
+	myID += myss.str();
 	if (opt_verbose)
 		std::cerr << "RBC: myID = " << myID << std::endl;
 	// assume maximum asynchronous t-resilience for RBC
@@ -722,7 +721,7 @@ void run_instance
 	else
 	{
 		// generate a non-shared DSA primary key, if s-resilience is zero
-		tmcg_mpz_ssrandomm_cache(cache, cache_mod, &cache_avail, dsa_x, vtmf->q);
+		tmcg_mpz_ssrandomm_cache(cache, cache_mod, cache_avail, dsa_x, vtmf->q);
 		tmcg_mpz_spowm(dsa_y, vtmf->g, dsa_x, vtmf->p);
 	}
 	// extract further public parameters for OpenPGP key structures
@@ -1519,101 +1518,6 @@ void run_instance
 			std::endl;
 	rbc->Sync(synctime);
 
-	// export generated public keys in OpenPGP armor format
-	std::stringstream pubfilename;
-	pubfilename << peers[whoami] << "_dkg-pub.asc";
-	armor = "", all.clear();
-	all.insert(all.end(), pub.begin(), pub.end());
-	all.insert(all.end(), uid.begin(), uid.end());
-	all.insert(all.end(), uidsig.begin(), uidsig.end());
-	all.insert(all.end(), sub.begin(), sub.end());
-	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
-	if (opt_verbose > 1)
-		std::cout << armor << std::endl;
-	std::ofstream pubofs((pubfilename.str()).c_str(), std::ofstream::out);
-	if (!pubofs.good())
-	{
-		std::cerr << "ERROR: P_" << whoami <<
-			": opening public key file failed" << std::endl;
-		delete dkg, delete dss;
-		delete rbc, delete vtmf, delete aiou, delete aiou2;
-		exit(-1);
-	}
-	pubofs << armor;
-	if (!pubofs.good())
-	{
-		std::cerr << "ERROR: P_" << whoami << 
-			": writing public key file failed" << std::endl;
-		delete dkg, delete dss;
-		delete rbc, delete vtmf, delete aiou, delete aiou2;
-		exit(-1);
-	}
-	pubofs.close();
-
-	// export generated private keys in OpenPGP armor format
-	std::stringstream secfilename;
-	secfilename << peers[whoami] << "_dkg-sec.asc";
-	armor = "", all.clear();
-	all.insert(all.end(), sec.begin(), sec.end());
-	all.insert(all.end(), uid.begin(), uid.end());
-	all.insert(all.end(), uidsig.begin(), uidsig.end());
-	all.insert(all.end(), ssb.begin(), ssb.end());
-	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::
-		ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
-	if (opt_verbose > 1)
-		std::cout << armor << std::endl;
-	if (!create_strict_permissions((secfilename.str()).c_str()))
-	{
-		if (errno == EEXIST)
-		{
-			if (!check_strict_permissions((secfilename.str()).c_str()))
-			{
-				std::cerr << "WARNING: weak permissions of existing private" <<
-					" key file detected" << std::endl;
-				if (!set_strict_permissions((secfilename.str()).c_str()))
-				{
-					std::cerr << "ERROR: P_" << whoami << ": setting" <<
-						" permissions for private key file failed" << std::endl;
-					delete dkg, delete dss;
-					delete rbc, delete vtmf, delete aiou, delete aiou2;
-					exit(-1);
-				}
-			}
-			std::cerr << "WARNING: existing private key file have been" <<
-				" overwritten" << std::endl;
-		}
-		else
-		{
-			std::cerr << "ERROR: P_" << whoami << ": creating private key" <<
-				" file failed" << std::endl;
-			delete dkg, delete dss;
-			delete rbc, delete vtmf, delete aiou, delete aiou2;
-			exit(-1);
-		}
-	}
-	std::ofstream secofs((secfilename.str()).c_str(), std::ofstream::out);
-	if (!secofs.good())
-	{
-		std::cerr << "ERROR: P_" << whoami << ": opening private key file" <<
-			" failed" << std::endl;
-		delete dkg, delete dss;
-		delete rbc, delete vtmf, delete aiou, delete aiou2;
-		exit(-1);
-	}
-	secofs << armor;
-	if (!secofs.good())
-	{
-		std::cerr << "ERROR: P_" << whoami << ": writing private key file" <<
-			" failed" << std::endl;
-		delete dkg, delete dss;
-		delete rbc, delete vtmf, delete aiou, delete aiou2;
-		exit(-1);
-	}
-	secofs.close();
-
 	// release DKG
 	delete dkg;
 
@@ -1646,6 +1550,64 @@ void run_instance
 
 	// release asynchronous unicast and broadcast
 	delete aiou, delete aiou2;
+
+	// export generated public keys in OpenPGP armor format
+	std::stringstream pubfilename;
+	pubfilename << peers[whoami] << "_dkg-pub.asc";
+	armor = "", all.clear();
+	all.insert(all.end(), pub.begin(), pub.end());
+	all.insert(all.end(), uid.begin(), uid.end());
+	all.insert(all.end(), uidsig.begin(), uidsig.end());
+	all.insert(all.end(), sub.begin(), sub.end());
+	all.insert(all.end(), subsig.begin(), subsig.end());
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
+	if (opt_verbose > 1)
+		std::cout << armor << std::endl;
+	if (!write_key_file(pubfilename.str(), armor))
+		exit(-1);
+
+	// export generated private keys in OpenPGP armor format
+	std::stringstream secfilename;
+	secfilename << peers[whoami] << "_dkg-sec.asc";
+	std::string sfilename = secfilename.str();
+	armor = "", all.clear();
+	all.insert(all.end(), sec.begin(), sec.end());
+	all.insert(all.end(), uid.begin(), uid.end());
+	all.insert(all.end(), uidsig.begin(), uidsig.end());
+	all.insert(all.end(), ssb.begin(), ssb.end());
+	all.insert(all.end(), subsig.begin(), subsig.end());
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
+	if (opt_verbose > 1)
+		std::cout << armor << std::endl;
+	if (!create_strict_permissions(sfilename))
+	{
+		if (errno == EEXIST)
+		{
+			if (!check_strict_permissions(sfilename))
+			{
+				std::cerr << "WARNING: weak permissions of existing private" <<
+					" key file detected" << std::endl;
+				if (!set_strict_permissions(sfilename))
+				{
+					std::cerr << "ERROR: P_" << whoami << ": setting" <<
+						" permissions for private key file failed" << std::endl;
+					exit(-1);
+				}
+			}
+			std::cerr << "WARNING: existing private key file have been" <<
+				" overwritten" << std::endl;
+		}
+		else
+		{
+			std::cerr << "ERROR: P_" << whoami << ": creating private key" <<
+				" file failed" << std::endl;
+			exit(-1);
+		}
+	}
+	if (!write_key_file(sfilename, armor))
+		exit(-1);
 }
 
 bool fips_verify
@@ -1984,7 +1946,7 @@ int main
 		GNUNET_GETOPT_option_help(about),
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
-			"TIME",
+			"INTEGER",
 			"expiration time of generated keys in seconds",
 			&gnunet_opt_keyexptime
 		),
@@ -2034,13 +1996,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of key generation protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -2155,7 +2117,7 @@ int main
 				std::cout << "Arguments mandatory for long options are also" <<
 					" mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
-				std::cout << "  -e TIME        expiration time of generated" <<
+				std::cout << "  -e INTEGER     expiration time of generated" <<
 					" keys in seconds" << std::endl;
 				std::cout << "  -g STRING      common reference string that" <<
 					" defines underlying DDH-hard group" << std::endl;
@@ -2173,7 +2135,7 @@ int main
 					std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" <<
 					std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point" <<
+				std::cout << "  -W INTEGER     timeout for point-to-point" <<
 					" messages in minutes" << std::endl;
 				std::cout << "  -y, --yaot     yet another OpenPGP tool" <<
 					std::endl;
@@ -2561,10 +2523,9 @@ int main
 	std::cerr << "   Please use other programs, move the mouse, and type on" <<
 		" your keyboard: " << std::endl;
 	if (opt_y)
-		tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail, 2,
-			fips_q);
+		tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, cache_avail, 2, fips_q);
 	else
-		tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail,
+		tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, cache_avail,
 			((2 * (S + 1)) + (2 * (T + 1))), fips_q);
 	std::cerr << "Thank you!" << std::endl;
 	mpz_clear(fips_p), mpz_clear(fips_q), mpz_clear(fips_g);
@@ -2593,7 +2554,7 @@ int main
 		tcpip_close();
 		tcpip_done();
 		// release cache
-		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
 		// finish
 		return ret;
 	}
@@ -2602,7 +2563,7 @@ int main
 		// run as replacement for GnuPG et al. (yet-another-openpgp-tool)
 		run_instance(0, time(NULL), opt_e, 0);
 		// release cache
-		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
 		return ret;
 	}
 
@@ -2610,7 +2571,7 @@ int main
 	static const struct GNUNET_GETOPT_CommandLineOption myoptions[] = {
 		GNUNET_GETOPT_option_uint('e',
 			"expiration",
-			"TIME",
+			"INTEGER",
 			"expiration time of generated keys in seconds",
 			&gnunet_opt_keyexptime
 		),
@@ -2657,13 +2618,13 @@ int main
 		),
 		GNUNET_GETOPT_option_uint('w',
 			"wait",
-			"TIME",
+			"INTEGER",
 			"minutes to wait until start of key generation protocol",
 			&gnunet_opt_wait
 		),
 		GNUNET_GETOPT_option_uint('W',
 			"aiou-timeout",
-			"TIME",
+			"INTEGER",
 			"timeout for point-to-point messages in minutes",
 			&gnunet_opt_W
 		),
@@ -2684,7 +2645,7 @@ int main
 		argv[0]);
 //	GNUNET_free((void *) argv);
 	// release cache
-	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
 	// finish
 	if (ret == GNUNET_OK)
 		return 0;
@@ -2758,7 +2719,7 @@ int main
 	}
 	
 	// release cache
-	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, cache_avail);
 	// finish
 	return ret;
 }
